"""This module provides a command-line interface to the ensemble themis."""
# pylint: disable=too-many-lines

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import argparse
import sys
import os
import json
import traceback


if __name__ == "__main__":
    sys.path.append(os.path.abspath(__file__).rsplit(os.sep, 2)[0])


# pylint: disable=wrong-import-position
import themis
from themis import utils
from themis.versions import reprlib
from themis import database
from themis import runtime


def get_themis(args):
    """Get the Themis object from the given setup directory or exit."""
    try:
        return themis.Themis(args.setup_dir)
    except ValueError as val_err:
        sys.exit("Invalid value given for the `--setup-dir` argument: " + str(val_err))


def add_subparser_aliases(subparsers, aliases, *args, **kwargs):
    """Add aliases to subcommands."""
    if sys.version_info.major > 2:
        return subparsers.add_parser(*args, aliases=aliases, **kwargs)
    return subparsers.add_parser(*args, **kwargs)


def add_verbose(parser):
    """Add a verbose argument to a parser."""
    parser.add_argument(
        "-v",
        "--verbose",
        action="count",
        default=0,
        help="Write more information to stdout",
    )


def add_parallel(parser):
    """Add a parallel argument to a parser."""
    parser.add_argument(
        "--parallel",
        type=int,
        default=None,
        help="reserve N cores for Themis to run in parallel on",
        metavar="N",
    )


def add_multiple_option(parser):
    """Add a multiple argument to a parser."""
    parser.add_argument(
        "--allow-multiple",
        action="store_true",
        help=(
            "Allow multiple concurrent executions. "
            "This flag must be set on all concurrent executions."
        ),
    )


def composite_create_common_args(parser):
    """Arguments common to 'create' and 'create-composite'."""
    parser.add_argument(
        "--flux",
        nargs="?",
        const=True,
        default=False,
        help=(
            "Use the Flux resource manager, optionally passing "
            "a path to the installation to use"
        ),
        metavar="FLUX_PATH",
    )
    parser.add_argument(
        "--max-restarts",
        type=int,
        default=0,
        metavar="N",
        help=(
            "The number of times to restart a run upon failure "
            "(i.e. a nonzero returncode)."
        ),
    )
    parser.add_argument(
        "--overwrite",
        action="store_true",
        help=("Create a new Themis ensemble, force-removing any existing one"),
    )
    parser.add_argument("--interface", help=argparse.SUPPRESS)
    utils.add_common_creation_args(parser)
    parser.add_argument(
        "--abort-on",
        nargs="+",
        type=int,
        help=(
            "One or more OS process return codes (positive integers). "
            "If `application` exits with one of these return codes, "
            "the run will be marked as 'aborted' and will not be restarted."
        ),
    )


def handle_status_subcommand(args):
    """Display the run IDs of runs with certain statuses."""
    themis_obj = get_themis(args)
    status_enums = {themis_obj.translate_enum(name) for name in args.status_names}
    if args.count:
        print(themis_obj.count_by_status(*status_enums))
    else:
        # print the run IDs by adding a space between every element
        print(" ".join((str(i) for i in themis_obj.filter_by_status(*status_enums))))


def setup_status_parser(subparsers):
    """Setup the parser for the status subcommand."""
    status_choices = [
        themis.Themis.translate_enum(enum) for enum in themis.Themis.STATUS_ENUMS
    ]
    status_choices.append([])
    status_parser = subparsers.add_parser(
        "status",
        help="Filter runs by status",
        description=(
            "Filter runs by status; return the IDs of runs with the "
            "specified statuses. If no statuses are specified, return the IDs "
            "of all runs."
        ),
    )
    status_parser.add_argument(
        "status_names",
        choices=status_choices,
        nargs="*",
        type=lambda x: str(x).lower(),
        help="The status filters to apply",
    )
    status_parser.add_argument(
        "-c",
        "--count",
        action="store_true",
        help="count the number of runs instead of returning their ids",
    )
    status_parser.set_defaults(handler=handle_status_subcommand)


def handle_kill_subcommand(args):
    """Kill runs."""
    themis_obj = get_themis(args)
    themis_obj.kill_runs(args.run_ids)


def setup_kill_parser(subparsers):
    """Setup the parser for the kill subcommand."""
    kill_parser = add_subparser_aliases(
        subparsers,
        ["dequeue"],
        "kill",
        help="Remove runs from the queue",
        description=(
            "Remove runs from the queue, so they will not be executed. "
            "Has no effect on completed runs."
        ),
    )
    kill_parser.add_argument(
        "run_ids", type=int, nargs="+", help="ID of a run to kill", metavar="ID"
    )
    kill_parser.set_defaults(handler=handle_kill_subcommand)


def _repr_unknown(repr_instance, obj, maxsize):
    """Compactly repr an unknown, possibly large and nested, object."""
    old_maxstring = repr_instance.maxstring
    repr_instance.maxstring = maxsize
    return_val = repr_instance.repr(repr_instance.repr(obj)).strip("\"'")
    repr_instance.maxstring = old_maxstring
    return return_val


def _repr_str(repr_instance, str_to_repr, maxsize):
    old_maxstring = repr_instance.maxstring
    repr_instance.maxstring = maxsize
    return_val = repr_instance.repr(str_to_repr).strip("\"'")
    repr_instance.maxstring = old_maxstring
    return return_val


def handle_display_subcommand(args):  # pylint: disable=too-many-locals
    """Print a table of data about the ensemble.

    Certain columns are only displayed optionally.
    """
    themis_obj = get_themis(args)
    id_args_repr = reprlib.Repr()
    id_args_repr.maxstring = 15
    id_args_repr.maxlong = 9
    resource_repr = reprlib.Repr()
    resource_repr.maxlong = 5
    custom_repr = reprlib.Repr()
    run_dicts = themis_obj.runs(range(args.lowerbound, args.upperbound))
    sample_header = None
    sample_cols = ()
    run_dir_names = database.get_app_spec(themis_obj.setup_dir)["run_dir_names"]
    if not run_dicts:
        print("No runs found in that range")
        return
    if args.all:
        args.directory = max(args.directory, args.all)
        args.arguments = max(args.arguments, args.all)
        args.sample = max(args.sample, args.all)
        args.result = max(args.result, args.all)
    format_string = "|{run_id:^7}|{step:^6}|{status:^12}|"
    if args.directory:
        # ensure column has at least minimum width
        args.directory = max(args.directory, len("Directory"))
        format_string += "{cwd:^" + str(args.directory) + "}|"
    if args.arguments:
        # ensure column has at least minimum width
        args.arguments = max(args.arguments, len("Args"))
        format_string += "{args:^" + str(args.arguments) + "}|"
    if args.resources or args.all:
        format_string += "{tasks:^5}|{cores:^5}|{gpus:^4}|"
    if args.sample:
        # ensure column has at least minimum width
        args.sample = max(args.sample, 5)
        # get the first sample and use its keys as the columns
        sample_cols = sorted(next(iter(run_dicts.values())).sample.keys(), key=str)[
            : args.colcount
        ]
        format_string += "{sample}"
        sample_header = "".join(
            [
                ("{:^" + str(args.sample) + "}|").format(
                    _repr_unknown(custom_repr, col, args.sample)
                )
                for col in sample_cols
            ]
        )
    if args.result:
        # ensure column has at least minimum width
        args.result = max(args.result, len("Result"))
        format_string += "{result:^" + str(args.result) + "}|"
    header_row = format_string.format(
        run_id="Run",
        status="Status",
        step="Step",
        cwd="Directory",
        args="Args",
        tasks="Tasks",
        cores="Cores",
        gpus="GPUs",
        sample=sample_header,
        result="Result",
    )
    line_row = len(header_row) * "-"
    print("\n" + line_row + "\n" + header_row + "\n" + line_row)
    for run_id, run in run_dicts.items():
        for i, step in enumerate(run.steps):
            if i == 0:
                status = themis_obj.translate_enum(run.status)
                result = _repr_unknown(id_args_repr, run.result, args.result)
                run_id_display = id_args_repr.repr(run_id)
                get_sample_val = lambda sample, key: _repr_unknown(
                    custom_repr, sample.get(key, None), args.sample
                )
            else:
                run_id_display = status = result = ""
                get_sample_val = lambda _1, _2: ""  # return empty string
            print(
                format_string.format(
                    step=i,
                    run_id=run_id_display,
                    cwd=_repr_str(
                        custom_repr,
                        os.path.relpath(
                            utils.get_run_dir(run_id, run_dir_names, run.sample)
                        ),
                        args.directory,
                    ),
                    args=_repr_unknown(custom_repr, step.args, args.arguments),
                    tasks=resource_repr.repr(step.tasks),
                    cores=resource_repr.repr(step.cores_per_task * step.tasks),
                    gpus=resource_repr.repr(step.gpus_per_task * step.tasks),
                    sample="".join(
                        [
                            ("{:^" + str(args.sample) + "}|").format(
                                get_sample_val(run.sample, key)
                            )
                            for key in sample_cols
                        ]
                    ),
                    status=status,
                    result=result,
                )
            )
    print(line_row + "\n")


def setup_display_parser(subparsers):
    """Setup the parser for the display subcommand."""
    display_parser = subparsers.add_parser(
        "display",
        help="Print information about runs",
        description=(
            "Print a table of information about runs. By default, only prints out "
            "the status of each run. Some entries in the table will be cut down if "
            "they become too long; use the `write` subcommand "
            "for a complete representation."
        ),
    )
    display_parser.add_argument(
        "lowerbound",
        type=int,
        help="The lower bound of the runs to display, inclusive",
    )
    display_parser.add_argument(
        "upperbound",
        type=int,
        help="The upper bound of the runs to display, exclusive",
    )
    display_parser.add_argument(
        "-a",
        "--arguments",
        nargs="?",
        type=int,
        default=0,
        const=15,
        help=(
            "Display (with a given column width) the command-line "
            "arguments for each run"
        ),
    )
    display_parser.add_argument(
        "-r",
        "--result",
        nargs="?",
        type=int,
        default=0,
        const=15,
        help="Display (with a given column width) the result for each run",
    )
    display_parser.add_argument(
        "-t",
        "--resources",
        action="store_true",
        help=(
            "Display the resources for each run, in terms of total tasks, cores, "
            "and gpus"
        ),
    )
    display_parser.add_argument(
        "-s",
        "--sample",
        nargs="?",
        type=int,
        default=0,
        const=12,
        help=(
            "Display (with a given column width) the sample for each run. "
            "Use the `-c` option to change the number of columns displayed."
        ),
    )
    display_parser.add_argument(
        "-c",
        "--colcount",
        type=int,
        default=4,
        help="Set the max number of sample columns to display. Defaults to 4.",
        metavar="N",
    )
    display_parser.add_argument(
        "-d",
        "--directory",
        help="Display (with a given column width) the working directory for each run",
        nargs="?",
        type=int,
        default=0,
        const=15,
    )
    display_parser.add_argument(
        "--all",
        nargs="?",
        type=int,
        default=0,
        const=15,
        help="Display all the available information about each run",
    )
    display_parser.set_defaults(handler=handle_display_subcommand)


def handle_post_subcommand(args):
    """Call the `themis.Themis.call_post_run` method."""
    themis_obj = get_themis(args)
    restart = themis.Themis(themis_obj.setup_dir)
    if not args.run_ids:
        args.run_ids = themis_obj.filter_by_status()
    for run_id in args.run_ids:
        post_run_result = restart.call_post_run(run_id)
        if args.verbose > 0:
            print("Result of post_run {}: {}".format(run_id, post_run_result))
        if args.save:
            themis_obj.set_result(run_id, post_run_result)


def setup_post_parser(subparsers):
    """Setup the parser for the post subcommand."""
    post_parser = subparsers.add_parser(
        "post",
        help="Call the `post_run` application interface function",
        description=(
            "Call the `post_run` application interface function for a number of runs. "
            "Has no effect if the `post_run` function is not defined. "
            "Unless specific run IDs are passed, call `post_run` for the whole "
            "ensemble."
        ),
    )
    post_parser.add_argument(
        "run_ids",
        nargs="*",
        type=int,
        help="the run IDs of the runs to call `post_run` for",
        metavar="ID",
    )
    post_parser.add_argument(
        "-s",
        "--save",
        action="store_true",
        help=(
            "Save the result internally, overwriting the result currently stored "
            "for the run. The new result will appear when fetching data about the "
            "ensemble."
        ),
    )
    add_verbose(post_parser)
    post_parser.set_defaults(handler=handle_post_subcommand)


def handle_dryrun_subcommand(args):
    """Call the `themis.Themis.dry_run` method."""
    themis_obj = get_themis(args)
    themis.Themis(themis_obj.setup_dir).dry_run(*args.run_ids, verbosity=args.verbose)


def setup_dryrun_parser(subparsers):
    """Setup the parser for the dryrun subcommand."""
    dryrun_parser = subparsers.add_parser(
        "dryrun",
        help="Set up runs, but do not execute",
        description=(
            "Dry-run an ensemble or specific runs within an ensemble. "
            "Create the runs' directories and populate them. "
            "Call the `prep_run` application interface function, if it exists. "
            "Unless specific run IDs are passed, dry run the whole ensemble. "
            "The dry-runs are done in serial, so it is safe to call this from "
            "the login node of a HPC cluster."
        ),
    )
    dryrun_parser.add_argument(
        "run_ids", nargs="*", type=int, help="run ID to dry-run", metavar="ID"
    )
    add_verbose(dryrun_parser)
    dryrun_parser.set_defaults(handler=handle_dryrun_subcommand)


def get_runs_from_parameterfile(args):
    """Return a list of Run objects as read from a CSV"""
    if args.parameterfile is None:
        return None
    samples = utils.read_csv(args.parameterfile)
    if args.vary_all:
        return [
            themis.Run(
                sample=sample,
                args=sample.get("args", args.args),
                tasks=sample.get("tasks", args.ntasks),
                cores_per_task=sample.get("cores_per_task", args.cores_per_task),
                gpus_per_task=sample.get("gpus_per_task", args.gpus_per_task),
            )
            for sample in samples
        ]
    return [
        themis.Run(
            sample=sample,
            args=args.args,
            tasks=args.ntasks,
            cores_per_task=args.cores_per_task,
            gpus_per_task=args.gpus_per_task,
        )
        for sample in samples
    ]


def add_related_parameterfile_args(parser):
    """Add arguments relating to

    Doesn't add the parameterfile argument itself.
    """
    parser.add_argument(
        "-a",
        "--args",
        type=str,
        default=None,
        help="Command-line arguments to pass to the application",
    )
    parser.add_argument(
        "-n", "--ntasks", type=int, default=1, help="MPI tasks per run", metavar="N"
    )
    parser.add_argument(
        "-c",
        "--cores-per-task",
        type=int,
        default=1,
        help="Cores per MPI task",
        metavar="N",
    )
    parser.add_argument(
        "-g",
        "--gpus-per-task",
        type=int,
        default=0,
        help="GPUs per MPI task",
        metavar="N",
    )
    parser.add_argument(
        "--vary-all",
        action="store_true",
        help=(
            "Look for `args`, `tasks`, `cores_per_task`, and `gpus_per_task` entries "
            "in `parameterfile` and use those values instead of `--tasks`, `--args`, "
            "etc."
        ),
    )


def create_ensemble(args, runs):
    """Create an ensemble given runs and args."""
    if themis.Themis.exists(args.setup_dir) and not args.overwrite:
        print(
            "Themis setup exists in {!r} and `--overwrite` flag not given, "
            "doing nothing...".format(args.setup_dir)
        )
    else:
        themis.Themis.create_overwrite(
            application=args.application,
            runs=runs,
            setup_dir=args.setup_dir,
            run_parse=args.run_parse,
            run_symlink=args.run_symlink,
            run_copy=args.run_copy,
            run_dir_names=args.run_dir_names,
            app_is_batch_script=not args.no_batch_script,
            use_flux=args.flux,
            max_restarts=args.max_restarts,
            app_interface=args.interface,
            abort_on=args.abort_on,
        )


def handle_create_subcommand(args):
    """Create a new ensemble from an application and a samples file."""
    runs = get_runs_from_parameterfile(args)
    create_ensemble(args, runs)


def setup_create_parser(subparsers):
    """Setup the parser for the create subcommand."""
    create_parser = subparsers.add_parser(
        "create",
        help="Create a new Themis ensemble",
        description=(
            "Create a new Themis ensemble from an application and a file defining the "
            "parameters for each run. Resources and command-line arguments are "
            "assumed to be constant across runs unless `--vary-all` is set. "
            "Setup files are put into the `--setup-dir` directory. "
            "If there are already setup files in `--setup-dir`, nothing is done "
            "unless the `--overwrite` flag is given."
        ),
    )
    create_parser.add_argument(
        "application", type=str, help="Path to the application to execute in parallel"
    )
    create_parser.add_argument(
        "--no-batch-script",
        action="store_true",
        help=(
            "Add this flag if `application` is NOT a batch "
            "script or any other sort of script that launches other "
            "applications."
        ),
    )
    composite_create_common_args(create_parser)
    add_related_parameterfile_args(create_parser)
    utils.add_parameterfile_arg(create_parser, "?")
    create_parser.set_defaults(handler=handle_create_subcommand)


def build_composite_runs(parameterfile, stepfile):
    """Build composite runs from two CSVs."""
    samples = utils.read_csv(parameterfile)
    steps = [themis.Step(**step) for step in utils.read_csv(stepfile)]
    return [themis.CompositeRun(sample, steps) for sample in samples]


def add_stepfile_arg(parser):
    """Add a stepfile argument to a parser."""
    parser.add_argument(
        "stepfile",
        help=(
            "Path to the CSV file defining the steps for each run. "
            "Each row must define a single step. "
            "The accepted columns (you may specify only a subset) are "
            "args, tasks, cores_per_task, gpus_per_task, timeout, "
            "batch_script, and cwd."
        ),
    )


def handle_create_comp_subcommand(args):
    """Create a new ensemble of composite runs."""
    runs = build_composite_runs(args.parameterfile, args.stepfile)
    args.no_batch_script = False
    args.application = None
    create_ensemble(args, runs)


def setup_create_composite_parser(subparsers):
    """Setup the parser for the create-composite subcommand."""
    create_parser = subparsers.add_parser(
        "create-composite",
        help="Create a new composite Themis ensemble",
        description=(
            "Create a new Themis ensemble from a file defining the samples for each "
            "run and a file defining the steps for each run. "
            "Steps are assumed to be constant across all runs. "
            "Setup files are put into the `--setup-dir` directory. "
            "If there are already setup files in `--setup-dir`, nothing is done "
            "unless the `--overwrite` flag is given."
        ),
    )
    composite_create_common_args(create_parser)
    utils.add_parameterfile_arg(create_parser, None)
    add_stepfile_arg(create_parser)
    create_parser.set_defaults(handler=handle_create_comp_subcommand)


def handle_add_subcommand(args):
    """Handle the `add_runs` subcommand."""
    themis_obj = get_themis(args)
    runs = get_runs_from_parameterfile(args)
    themis_obj.add_runs(runs)


def setup_add_parser(subparsers):
    """Add parser for add runs subcommand"""
    parser = subparsers.add_parser(
        "add",
        help="Add new runs",
        description=(
            "Add runs to an ensemble. Resources and command-line arguments are "
            "assumed to be constant across runs unless `--vary-all` is set. "
        ),
    )
    utils.add_parameterfile_arg(parser)
    add_related_parameterfile_args(parser)
    parser.set_defaults(handler=handle_add_subcommand)


def handle_add_composite_command(args):
    """Build composite runs and add them to the ensemble."""
    themis_obj = get_themis(args)
    runs = build_composite_runs(args.parameterfile, args.stepfile)
    themis_obj.add_runs(runs)


def setup_add_composite_parser(subparsers):
    """Add a parser for adding composite runs."""
    parser = subparsers.add_parser(
        "add-composite",
        help="Add new composite (multi-step) runs",
        description=(
            "Add composite (multi-step) runs to an ensemble. Steps are assumed "
            "to be constant across all runs."
        ),
    )
    utils.add_parameterfile_arg(parser)
    add_stepfile_arg(parser)
    parser.set_defaults(handler=handle_add_composite_command)


def handle_progress_subcommand(args):
    """Print a summary progress line."""
    themis_obj = get_themis(args)
    completed, total = themis_obj.progress()
    if args.verbose > 0:
        for enum in themis_obj.STATUS_ENUMS:
            translation = themis_obj.translate_enum(enum)
            print(
                translation.capitalize()
                + " runs: "
                + str(themis_obj.count_by_status(enum))
            )
    print()
    utils.print_progress_bar(completed, total, suffix="Complete", length=35)
    # print newline to clear, plus another newline to add an actual blank line
    # one newline is implicit in the call to print
    print("\n")


def setup_progress_parser(subparsers):
    """Setup the parser for the progress subcommand."""
    progress_parser = subparsers.add_parser(
        "progress",
        help="Print out a short progress summary",
        description=(
            "Print out a short summary of an ensemble's progress. "
            "Shows the percentage and number of completed runs."
        ),
    )
    add_verbose(progress_parser)
    progress_parser.set_defaults(handler=handle_progress_subcommand)


def handle_requeue_subcommand(args):
    """Restart runs."""
    get_themis(args).requeue_runs(args.run_ids, args.hard)


def setup_requeue_parser(subparsers):
    """Setup the parser for the restart subcommand."""
    requeue_parser = add_subparser_aliases(
        subparsers,
        ["requeue"],
        "restart",
        help="Mark runs as eligible for execution",
        description=(
            "Mark runs as eligible for restart. The specified runs will be given "
            " the 'queued' status and will be executed "
            "by Themis the next time it starts, "
            "or immediately if Themis is currently active."
        ),
    )
    requeue_parser.add_argument(
        "run_ids", type=int, nargs="+", help="ID of a run to restart", metavar="ID"
    )
    requeue_parser.add_argument(
        "--hard",
        action="store_true",
        help=(
            "If set, reset the runs back to step 0. "
            "Otherwise, leave the runs' progress unchanged."
        ),
    )
    requeue_parser.set_defaults(handler=handle_requeue_subcommand)


def handle_execute_alloc_subcommand(args):
    """Execute the ensemble and print the job ID."""
    themis_obj = get_themis(args)
    print(
        "Batch job ID is "
        + str(
            themis_obj.execute_alloc(
                nodes=args.nodes,
                timeout=args.timeout,
                partition=args.partition,
                bank=args.bank,
                name=args.name,
                repeats=args.repeats,
                parallelism=args.parallel,
                early_stop=args.early_stop,
                allow_multiple=args.allow_multiple,
            )
        )
    )


def setup_execute_alloc_parser(subparsers):
    """Setup the parser for the execute subcommand."""
    execute_parser = subparsers.add_parser(
        "execute-alloc",
        help="Execute an ensemble in a new batch allocation",
        description=(
            "Request a new batch allocation and launch an ensemble within it. "
            "Finish any incomplete runs. Already-completed runs will not be restarted."
        ),
    )
    execute_parser.add_argument(
        "-N", "--nodes", help="the number of nodes to request", type=int, default=1
    )
    execute_parser.add_argument(
        "-p",
        "--partition",
        help="the partition to request the allocation in",
        type=str,
        metavar="NAME",
        default=None,
    )
    execute_parser.add_argument(
        "-b",
        "--bank",
        help="the bank to charge the allocation to",
        type=str,
        metavar="NAME",
        default=None,
    )
    execute_parser.add_argument(
        "-m",
        "--name",
        help="the name to give the allocation",
        type=str,
        default="themis",
    )
    execute_parser.add_argument(
        "-t",
        "--timeout",
        help="the timeout on the allocation in minutes",
        type=int,
        default=60,
        metavar="N",
    )
    execute_parser.add_argument(
        "-r",
        "--repeats",
        help=(
            "The number of times to replicate the allocation if time expires but "
            "the ensemble is not yet complete. -1 allows infinite repeats."
        ),
        type=int,
        default=0,
        metavar="N",
    )
    execute_parser.add_argument(
        "--early-stop",
        help=(
            "The number of minutes before the allocation expires that Themis should "
            "stop submitting new jobs. Makes Themis's overall runtime equivalent to "
            "`--timeout` minus `--early-stop`. Useful if jobs pay attention to the "
            "remaining allocation time."
        ),
        type=int,
        default=0,
        metavar="N",
    )
    add_multiple_option(execute_parser)
    add_parallel(execute_parser)
    execute_parser.set_defaults(handler=handle_execute_alloc_subcommand)


def handle_execute_local_subcommand(args):
    """Handle execute-local subcommand."""
    get_themis(args).execute_local(
        args.block, parallelism=args.parallel, allow_multiple=args.allow_multiple
    )


def setup_execute_local_parser(subparsers):
    """Setup parser for the execute-local subcommand."""
    execute_parser = subparsers.add_parser(
        "execute-local",
        help="Execute an ensemble locally",
        description=(
            "Execute an ensemble locally (i.e. do not request an allocation); "
            "finish any incomplete runs. Already-completed runs will not be restarted."
        ),
    )
    execute_parser.add_argument(
        "-b",
        "--block",
        action="store_true",
        help=("Block and do not exit until the ensemble is complete."),
    )
    add_multiple_option(execute_parser)
    add_parallel(execute_parser)
    execute_parser.set_defaults(handler=handle_execute_local_subcommand)


def handle_write_subcommand(args):
    """Write data about the ensemble to a file."""
    themis_obj = get_themis(args)
    if args.format == "json":
        themis_obj.write_json(sys.stdout)
    elif args.format == "yaml":
        themis_obj.write_yaml(sys.stdout)
    elif args.format == "csv":
        themis_obj.write_csv(sys.stdout)


def setup_write_parser(subparsers):
    """Setup the parser for the write subcommand."""
    write_parser = subparsers.add_parser(
        "write",
        help="Write a complete set of data in a variety of formats",
        description=(
            "Write information about an ensemble, such as parameters, resources,"
            " and outputs, to a file in a variety of formats."
        ),
    )
    write_parser.add_argument(
        "format",
        help="the file format in which to write the data",
        choices=["yaml", "json", "csv"],
        type=lambda x: str(x).lower(),
    )
    write_parser.set_defaults(handler=handle_write_subcommand)


def handle_rundir_subcommand(args):
    """Handle rundir subcommand."""
    themis_obj = get_themis(args)
    run_dirs = themis_obj.run_dirs(args.nonexistent)
    if args.json:
        print(json.dumps(list(run_dir for _, run_dir in run_dirs)))
    else:
        for _, run_dir in run_dirs:
            print(run_dir)


def setup_rundir_parser(subparsers):
    """Setup parser for the rundirs subcommand."""
    rundir_parser = subparsers.add_parser(
        "rundirs",
        help="Print the working directory for each run",
        description="Print the working directory for each run",
    )
    rundir_parser.add_argument(
        "--json", action="store_true", help="Print the run directories as JSON"
    )
    rundir_parser.add_argument(
        "-n",
        "--nonexistent",
        action="store_true",
        help="Include directories which haven't yet been created",
    )
    rundir_parser.set_defaults(handler=handle_rundir_subcommand)


def handle_completion_subcommand(args):
    """Handle the completion subcommand."""
    get_themis(args).on_completion(
        args.args, stdout=args.stdout, stderr=args.stderr, cwd=args.cwd
    )


def setup_completion_parser(subparsers):
    """Setup parser for the completion subcommand."""
    completion_parser = subparsers.add_parser(
        "completion",
        help="Provide an executable for Themis to launch upon completion",
        description=(
            "Provide an executable and arguments for Themis to launch "
            "once the ensemble has completed. Only the most recent invocation "
            "of this subcommand is remembered; multiple executables are not "
            "supported."
        ),
    )
    completion_parser.add_argument(
        "args",
        help="The executable and any associated arguments.",
        nargs=argparse.REMAINDER,
    )
    completion_parser.add_argument(
        "--stdout",
        help=(
            "Path to redirect stdout. Default redirects "
            "stdout to one of Themis's log files."
        ),
    )
    completion_parser.add_argument(
        "--stderr",
        help=(
            "Path to redirect stderr. Default redirects "
            "stderr to one of Themis's log files."
        ),
    )
    completion_parser.add_argument(
        "--cwd",
        help=(
            "Working directory for the executable. "
            "Default is the current working directory."
        ),
        default=os.curdir,
    )
    completion_parser.set_defaults(handler=handle_completion_subcommand)


def handle_runtime_parse(args):
    """Handle `themis runtime parse` command."""
    runtime.parse_file(args.source, args.dest)


def handle_runtime_collect(args):
    """Handle `themis runtime collect` subcommand."""
    with open(args.source, "rb") as file_handle:
        runtime.set_result(file_handle.read())


def setup_runtime_parser(subparsers):
    """Setup the runtime subparsers."""
    runtime_parser = subparsers.add_parser(
        "runtime",
        help="Themis runtime operations",
        description=(
            "A set of utilities available to applications launched by Themis."
        ),
    )
    runtime_subparsers = runtime_parser.add_subparsers(
        title="subcommands", description="", dest="runtime_subcommand"
    )
    parse_subparser = runtime_subparsers.add_parser(
        "parse",
        help="Parse and token-replace a text file",
        description=(
            "Search a text file for Themis tokens and replace them "
            "with values associated with the current run."
        ),
    )
    parse_subparser.add_argument("source", help="The file to parse")
    parse_subparser.add_argument("dest", help="The destination of the parsed file")
    parse_subparser.set_defaults(handler=handle_runtime_parse)
    collect_subparser = runtime_subparsers.add_parser(
        "collect",
        help="Collect and store the contents of a file",
        description=(
            "Collect the contents of a file and store it as the "
            "result of the current run."
        ),
    )
    collect_subparser.add_argument("source", help="The file to collect and store")
    collect_subparser.set_defaults(handler=handle_runtime_collect)


def setup_parsers():
    """Return an argparse.ArgumentParser with all the options and subcommands."""
    parser = argparse.ArgumentParser(
        description=(
            "Interact with or create a Themis ensemble. Usage of this "
            "program is broken into subcommands; see the help message on "
            "each subcommand for specifics. All subcommands except `create` "
            "will fail if an existing Themis ensemble "
            "is not found. To create and run a new ensemble, use `create`, and then "
            "one of the `execute-` commands. The `dryrun` command is useful for "
            "checking that an ensemble is configured properly."
        )
    )
    parser.add_argument(
        "--setup-dir",
        default=utils.DEFAULT_SETUP_DIR,
        help=(
            "The path to the directory of the ensemble to interact with or create. "
            "Defaults to {0!r}".format(utils.DEFAULT_SETUP_DIR)
        ),
        metavar="DIR",
    )
    add_verbose(parser)
    parser.add_argument(
        "--version", action="version", version="%(prog)s " + str(themis.__version__)
    )
    subparsers = parser.add_subparsers(
        title="subcommands", description="", dest="subcommand"
    )
    setup_create_parser(subparsers)
    setup_create_composite_parser(subparsers)
    setup_dryrun_parser(subparsers)
    setup_execute_alloc_parser(subparsers)
    setup_execute_local_parser(subparsers)
    setup_add_parser(subparsers)
    setup_add_composite_parser(subparsers)
    setup_progress_parser(subparsers)
    setup_display_parser(subparsers)
    setup_status_parser(subparsers)
    setup_write_parser(subparsers)
    setup_kill_parser(subparsers)
    setup_requeue_parser(subparsers)
    setup_rundir_parser(subparsers)
    setup_post_parser(subparsers)
    setup_completion_parser(subparsers)
    setup_runtime_parser(subparsers)
    return parser


def main():
    """Run the command-line program."""
    parser = setup_parsers()
    args = parser.parse_args()
    if not hasattr(args, "handler"):  # no subcommand given
        parser.print_usage()
        return
    try:
        args.handler(args)
    except Exception as exc:  # pylint: disable=broad-except
        if args.verbose:
            traceback.print_exc()
            sys.exit(1)
        else:
            sys.exit(type(exc).__name__ + ": " + str(exc))


if __name__ == "__main__":
    main()



"""
This module defines an interface for submitting runs to be executed.

The interface consists of the function setup_scheduler, which returns a
Scheduler object. Scheduler objects have in turn their own interface consisting of
the methods submit, shutdown, free_runs, kill_runs, active_runs, done, and
save_progress. See the ThreadedAppRunBatcher class for detail.

Some terminology that should be used consistently throughout this module:

    A step is the execution of an application. This involves not just executing
        the executable, but also setting the working directory, arguments, and
        so on. bundle of information about the creation of an application. A step
        corresponds to a single `srun` or `flux mini run` etc.
    A run is a series of steps executed in order: step 1 followed by step 2 followed
        by step 3 ... and so on. Each run corresponds to a sample, and is uniquely
        identified by a run ID.
"""

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import os
import sys
import logging
import time
import threading
import functools
from collections import deque

import themis.utils
from themis import resource
from themis import backend
from themis.database.utils import EnsembleDatabase
from themis.backend.worker import prepper


LOGGER = logging.getLogger(__name__)


def setup_scheduler(app_spec, server_args, max_active_runs, do_prep, do_post):
    """Create and return a Scheduler object.

    For information on Scheduler objects and their interface, consult the
    ThreadedAppRunBatcher class.
    """
    resource_mgr = resource.identify_resource_manager(app_spec["resource_mgr"])
    LOGGER.debug("Using resource manager %s", resource_mgr)
    step_creator = _StepCreator(app_spec, server_args, do_prep, do_post)
    scheduler = JobSubmissionThread(
        max_active_runs,
        functools.partial(
            _create_run, step_creator, app_spec["max_restarts"], app_spec["abort_on"]
        ),
        resource_mgr.executor(app_spec["setup_dir"], server_args),
        app_spec["max_restarts"],
    )
    scheduler.start()
    return scheduler


def _create_run(step_creator, max_restarts, abort_on, run_info, executor):
    """Create and return a new _AppRun object that will execute a run.

    :param step_creator: the StepCreator instance to be used for creating the steps
        for each new run.
    :param max_restarts: the number of times a step of each run may be
        restarted
    """
    return _AppRun(
        run_info.run_id,
        run_info.status,
        max_restarts - run_info.restarts,
        step_creator(run_info),
        abort_on,
        executor,
    )


# pylint: disable=too-many-instance-attributes
class JobSubmissionThread(threading.Thread):
    """Thread that creates and manages runs.

    :param waited_queue: a collections.Deque holding runs that have completed
    :param run_creator: callable for creating runs
    :param max_active_runs: the maximum number of concurrent runs to create
    """

    def __init__(self, max_active_runs, run_creator, executor, max_restarts):
        threading.Thread.__init__(self)
        self.daemon = True
        self.name = type(self).__name__
        self.profiler = None
        self.__exit_event = threading.Event()
        self.__executor = executor  # constant
        self.__run_creator = run_creator  # constant
        self.__max_active_runs = max_active_runs  # constant
        self.__max_restarts = max_restarts  # constant
        self.__runs = {}  # thread-private
        self.__done_lock = threading.Lock()  # public and thread-safe
        self.__new_run_queue = deque()  # public and thread-safe
        self.__statuses_to_write = deque()  # public and thread-safe
        self.__runs_to_kill = deque()  # public and thread-safe
        self.__done = False  # writes should be protected by __done_lock
        self.__jobids_to_runs = {}  # thread-private

    def run(self):
        """Profile the _run method if an environment variable is set."""
        self.profiler = backend.get_profiler()
        backend.profile(self.profiler, self._run)

    def submit(self, new_run_infos):
        """Submit runs; return when all have been submitted."""
        for run_info in new_run_infos:
            self.__new_run_queue.append(run_info)
        if new_run_infos:
            with self.__done_lock:
                self.__done = False

    def kill_runs(self, run_ids):
        """Kill the runs given by run_ids, provided they are currently active.

        Adds the run IDs to a data structure the thread will read from.
        The actual killing is left to that thread.
        """
        for run_id in run_ids:
            self.__runs_to_kill.append(run_id)

    def count_pending_runs(self):
        """Return the number of runs which have not yet been started."""
        return len(self.__new_run_queue)

    def complete_existing_runs(self):
        """Block until the scheduler has completed all active runs.

        Submitted but not active runs will not be scheduled."""
        while self.__new_run_queue:
            try:
                self.__new_run_queue.popleft().run_id
            except IndexError:  # queue may be emptied by another thread
                pass
        LOGGER.debug(
            "Draining runs from %s", type(self).__name__,
        )
        while not self.done():
            time.sleep(0.5)
        LOGGER.debug("%s drained", type(self).__name__)

    def save_progress(self, ensemble_db):
        """Save progress to the database."""
        run_ids_to_save = []
        statuses_to_save = []
        while self.__statuses_to_write:
            app_run = self.__statuses_to_write.popleft()
            run_ids_to_save.append(app_run.run_id)
            statuses_to_save.append(
                (
                    app_run.state,
                    app_run.completion_state,
                    self.__max_restarts - app_run.remaining_restarts,
                )
            )
        completed_ids_and_status = [
            (run_id, status[1])
            for run_id, status in zip(run_ids_to_save, statuses_to_save)
        ]
        ensemble_db.set_run_status(run_ids_to_save, statuses_to_save)
        return completed_ids_and_status

    def done(self):
        """Return True if there is no more work to be done."""
        return self.__done

    def shutdown(self, ensemble_db):
        """Shut down the instance."""
        self.__exit_event.set()
        self.join()
        for run in self.__runs.values():
            self.__statuses_to_write.append(run)
        self.save_progress(ensemble_db)
        return self.profiler

    def _run(self):
        """Loop through runs, attempting to advance them to the next state.

        The loop will not end until self._shutdown is set to True,
        Which must be done externally--there is no internal trigger to stop.
        """
        while not self.__exit_event.is_set():
            while len(self.__runs) < self.__max_active_runs and self.__new_run_queue:
                run_info = self.__new_run_queue.popleft()
                new_run = self.__run_creator(run_info, self.__executor)
                self.__runs[run_info.run_id] = new_run
                self.__jobids_to_runs[new_run.job_id] = new_run
            run_update = self.__executor.wait(timeout=0.1)
            if run_update is not None:
                self.__send_update(*run_update)
            if not self.__runs:
                with self.__done_lock:
                    if not self.__new_run_queue:
                        self.__done = True
            # if there are runs to kill, kill them
            while self.__runs_to_kill:
                run_id_to_kill = self.__runs_to_kill.popleft()
                if run_id_to_kill in self.__runs:
                    self.__runs[run_id_to_kill].kill_async()

    def __send_update(self, job_id, exit_status):
        """Update a run with the completion of one of its jobs."""
        run_to_update = self.__jobids_to_runs.pop(job_id, None)
        if run_to_update is None:
            raise ValueError(
                "Job ID {} (exitcode {}) doesn't match anything".format(
                    job_id, exit_status
                )
            )
        if run_to_update.run_complete(exit_status):
            self.__statuses_to_write.append(run_to_update)
            del self.__runs[run_to_update.run_id]
        else:
            self.__jobids_to_runs[run_to_update.job_id] = run_to_update


# pylint: disable=too-few-public-methods
class _StepCreator(object):
    """Instances of this class create Step objects.

    Instances store information which is constant across runs;
    Then, to create steps for a run, only the run RunInfo object
    for that run needs to be passed in.

    :param app_spec: the app_spec dictionary for the ensemble.
    :param server_info: the sequence of arguments to be used for
        connecting to the server
    """

    _THIS_FILE_LOCATION = os.path.dirname(os.path.abspath(__file__))

    def __init__(self, app_spec, server_info, do_prep, do_post):
        self._cwd = os.getcwd()
        self._app_spec = app_spec
        self._server_info = server_info
        # optimizations: no point launching a worker if it has nothing to do
        self._do_post = do_post
        self._do_prep = do_prep

    def __call__(self, run_info):
        """Return an interable of steps for the run given by run_info."""
        steps = []
        run_dir = themis.utils.get_run_dir(
            run_info.run_id, self._app_spec["run_dir_names"], run_info.sample
        )
        if not os.path.isdir(run_dir):
            os.makedirs(run_dir)
        if self._do_prep:
            steps.append(self._worker_step(run_info.run_id, run_dir, "prepper.py"))
        else:
            prepper.preparation(
                None, run_dir, self._app_spec, run_info.sample, run_info.steps
            )
        for step in run_info.steps:
            steps.append(self._user_step(step, run_dir))
        if self._do_post:
            steps.append(self._worker_step(run_info.run_id, run_dir, "finisher.py"))
        return steps

    @staticmethod
    def _user_step(step, run_dir):
        """Create steps for users' applications"""
        # if step.cwd is relative, make it relative to run dir
        step.cwd = os.path.join(run_dir, step.cwd)
        if not os.path.isdir(step.cwd):
            os.makedirs(step.cwd)
        step.args[0] = themis.utils.get_application(
            step.args[0], run_dir, step.batch_script,
        )
        return (step, None)

    def _worker_step(self, run_id, run_dir, worker_name):
        """Create a worker task step. The step is assumed to require a single core."""
        args = [
            sys.executable,
            os.path.join(self._THIS_FILE_LOCATION, worker_name),
            str(run_id),
            self._cwd,
            "--server-args",
        ]
        args.extend(self._server_info)
        if worker_name == "prepper.py":
            message = "Prepping run {}".format(run_id)
        else:
            message = "Finishing run {}".format(run_id)
        return (themis.utils.Step(args, batch_script=False, cwd=run_dir), message)


# pylint: disable=too-many-instance-attributes
class _AppRun(object):
    """Instances of this class represent and execute a single run of an ensemble.

    A run consists of one or more steps. Steps are represented by Step objects
    (naturally) and are expected to have a certain set of attributes.

    Upon construction, instances of this class execute their steps in order.
    A step is restarted upon failure (up to max_restarts times), and a step is only
    launched if there are resources available.

    Instances will not progress their steps in the background; the advance method
    must be repeatedly called in order for runs to make any progress.

    :param run_id: the run ID for this run
    :param state: the state of the current run. Usually 0, but may be different if
        the current run started previously, failed, and is now being restarted.
    :param steps: a sequence of Step objects representing the steps making up this run
    :param launcher: the Launcher class to be used for launching steps in this run
    :param max_restarts: the number of times a step may be restarted before the run
        is terminated.
    """

    # pylint: disable=too-many-arguments
    def __init__(self, run_id, state, remaining_restarts, steps, abort_on, executor):
        self.run_id = run_id
        self.state = state
        self.completion_state = EnsembleDatabase.RUN_QUEUED
        self._steps = steps
        self.remaining_restarts = remaining_restarts
        self._abort_on = abort_on
        self._executor = executor
        self.job_id = None
        self._done = False
        self._killed = False
        self._attempt_execution()

    def __repr__(self):
        """Return string representation of constructor call"""
        return "{}({}, {})".format(type(self).__name__, self.run_id, self.state)

    def _attempt_execution(self):
        """Attempt to execute the next step of this run"""
        curr_step, message = self._steps[self.state]
        if message is None:
            LOGGER.info("Launching run %i", self.run_id)
        else:
            LOGGER.info(message)
        self.job_id = self._executor.submit(self.run_id, curr_step)

    def run_complete(self, returncode):
        """Take action given that the most recent job has completed."""
        if returncode == 0:
            self.state += 1
            # check if the job has finished the final stage
            if self.state == len(self._steps):
                self.completion_state = EnsembleDatabase.RUN_SUCCESS
                LOGGER.info("Run %i completed successfully", self.run_id)
                self._done = True
                return True
            # if it hasn't, advance it to the next stage
            self._attempt_execution()
            return False
        # step hit fatal error, abort
        if returncode in self._abort_on:
            LOGGER.info("Run %i aborted with return code %s", self.run_id, returncode)
            self._done = True
            self.completion_state = EnsembleDatabase.RUN_ABORTED
            return True
        # The step exited with errors. check whether to restart
        if self._killed:
            LOGGER.info("Run %i was killed", self.run_id)
            self._done = True
            self.completion_state = EnsembleDatabase.RUN_KILLED
            return True
        LOGGER.info(
            "Failure in run %i on step %i: returncode %s",
            self.run_id,
            self.state,
            returncode,
        )
        if self.remaining_restarts != 0:
            # if remaining_restarts was -1, restart forever
            self.remaining_restarts -= 1
            self._attempt_execution()
            return False
        LOGGER.info("Run %i has hit the restart limit", self.run_id)
        self.completion_state = EnsembleDatabase.RUN_FAILURE
        self._done = True
        return True

    def kill_async(self):
        """Kill the run."""
        if not self._done:
            self._done = True
            self._killed = True
            self._executor.kill(self.job_id)



# -*- coding: utf-8 -*-

"""
This module defines utility functions and classes.

The classes and functions may be used by multiple other modules within this package.
This module is not intended for use by client code.
"""

from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

import os
import shutil
import signal
import glob
import warnings
import csv
import functools
import shlex

from themis.versions import Sequence, csvargs


DEFAULT_RUN_DIR_NAMES = "runs{sep}{{run_id}}".format(sep=os.sep)
DEFAULT_SETUP_DIR = os.path.join(os.curdir, ".themis_setup")
URI_ENV_VAR = "THEMIS_URI"
RUNID_ENV_VAR = "THEMIS_RUNID"
SETUPDIR_ENV_VAR = "THEMIS_SETUPDIR"
EXECDIR_ENV_VAR = "THEMIS_EXECDIR"
URI_SPLITCHAR = ":"


class Step(object):
    """Instances of this class represent the execution of one application.

    Instances do not do anything themselves; they are meant to be passed into
    other methods and functions.

    Many attributes map directly to ``lrun``, ``srun``, and ``flux mini run``
    arguments.

    :param args: the application and its arguments, either as a list of strings
        or as a single string, e.g. ``"/bin/bash -c 'exit 1;'"``
        or ``["/bin/bash", "-c", "exit 1;"]``. If a single string is passed,
        it will be converted to a list via ``shlex.split(args)``. The first
        element of the Sequence should be the *absolute* path to the
        application to execute.
    :type args: str or Sequence of str
    :param tasks: total MPI tasks. Corresponds to srun, lrun, and flux "-n" options.
    :type tasks: int, optional
    :param cores_per_task: cores per task. Corresponds to srun,
        lrun, and flux "-c" options.
    :type cores_per_task: int, optional
    :param gpus_per_task: gpus per task. Corresponds to lrun and flux "-g" option.
    :type gpus_per_task: int, optional
    :param timeout: the max running time, in minutes, for this run; if the limit is
        exceeded, the run will be killed.
    :type timeout: int, optional
    :param batch_script: a boolean indicating whether the application given by
        ``args[0]`` is a batch script. If it is, it will be parsed for tokens and
        copied into ``cwd`` before execution.
    :type batch_script: bool
    :param cwd: the working directory for the application given by ``args[0]``.
        If the path is relative, it is treated as relative the run directory for
        the owning run. By default, the CWD is the run directory.
    :type cwd: str
    """

    __slots__ = (
        "args",
        "tasks",
        "cores_per_task",
        "gpus_per_task",
        "timeout",
        "batch_script",
        "cwd",
    )

    def __init__(  # pylint: disable=too-many-arguments
        self,
        args,
        tasks=1,
        cores_per_task=1,
        gpus_per_task=0,
        timeout=0,
        batch_script=True,
        cwd=os.curdir,
    ):
        if isinstance(args, str):
            self.args = shlex.split(args)
        else:
            self.args = list(args)
        self.args[0] = os.path.abspath(self.args[0])
        self.tasks = int(tasks)
        self.cores_per_task = int(cores_per_task)
        self.gpus_per_task = int(gpus_per_task)
        self.timeout = int(timeout)
        self.batch_script = bool(batch_script)
        self.cwd = cwd if isinstance(cwd, str) else str(cwd)

    def __repr__(self):
        """Return string representation of constructor call for this object"""
        return (
            "{cls}(args={args}, tasks={tasks}, cores_per_task={cores}, "
            "gpus_per_task={gpus}, timeout={timeout}, batch_script={bscript}, "
            "cwd={cwd})"
        ).format(
            cls=type(self).__name__,
            args=self.args,
            tasks=self.tasks,
            cores=self.cores_per_task,
            gpus=self.gpus_per_task,
            timeout=self.timeout,
            bscript=self.batch_script,
            cwd=self.cwd,
        )

    @staticmethod
    def encode(obj):
        """JSON Encoder for Step objects."""
        if isinstance(obj, Step):
            return {key: getattr(obj, key) for key in obj.__slots__}
        raise TypeError()


class CompositeRun(object):  # pylint: disable=too-few-public-methods
    """Instances of this class represent one run of a Themis ensemble.

    ``CompositeRun`` instances are the most general kind of run in an ensemble, and
    consist of two parts: a mapping defining arbitrary key-value pairs, and a sequence
    of ``Step`` objects.

    Instances do not do anything themselves; they are meant to be passed into
    other methods and functions.

    Many attributes map directly to ``lrun``, ``srun``, and ``flux mini run``
    arguments.

    :param sample: a mapping from sample labels to values, such as one produced
        by iterating through a ``Samples`` object.
        Generally, sample labels (the keys of the mapping) should be constant
        across all runs. The sample will be used for parsing
        :ref:`text files <text_replacement>`.
    :type sample: Mapping
    :param steps: nonempty sequence of ``themis.Step`` objects defining the
        execution instructions for the run. Each ``Step``
        object owned by the run will be executed in order: first ``steps[0]``,
        then ``steps[1]``, then ``steps[2]`` and so on.
    :type steps: Sequence
    """

    __slots__ = ("sample", "steps")

    def __init__(self, sample, steps):
        self.sample = sample if sample is not None else {}
        self.steps = steps
        if not isinstance(steps, Sequence):
            raise TypeError(
                "Expected sequence type for steps, got {!r}".format(
                    type(steps).__name__
                )
            )
        if not steps:
            raise ValueError("len(steps) must be > 0")


class Run(CompositeRun):  # pylint: disable=too-many-instance-attributes
    """Represents one simple run of a Themis ensemble.

    This class should be named ``SimpleRun`` but is not for backwards compatibility.

    The ``Run`` class is a restrictive, simplifying derivation of the
    ``CompositeRun`` class. ``Run`` instances support only a single ``Step``,
    i.e. only a single application makes up the run. Other limitations include:

    * All ``Run`` instances in an ensemble are assumed to share the same application.
      As a consequence, whether or not the application is a batch script is set on an
      ensemble-wide basis.
    * The working directory for each ``Run`` in an ensemble is assumed to be specified
      by the ``run_dir_names`` ensemble-wide constant.

    Instances do not do anything themselves; they are meant to be passed into
    other methods and functions.

    Many attributes map directly to ``lrun``, ``srun``, and ``flux mini run``
    arguments.

    :param sample: a mapping from sample labels to values, such as one produced
        by iterating through a ``Samples`` object.
        Generally, sample labels (the keys of the mapping) should be constant
        across all runs. The sample will be used for parsing
        :ref:`text files <text_replacement>`.
    :param args: the command-line arguments to pass to the application.
    :type args: str, optional
    :param tasks: total MPI tasks. Corresponds to srun, lrun, and flux "-n" options.
    :type tasks: int, optional
    :param cores_per_task: cores per task. Corresponds to srun,
        lrun, and flux "-c" options.
    :type cores_per_task: int, optional
    :param gpus_per_task: gpus per task. Corresponds to lrun and flux "-g" option.
    :type gpus_per_task: int, optional
    :param timeout: the max running time, in minutes, for this run; if the limit is
        exceeded, the run will be killed.
    :type timeout: int, optional
    """

    __slots__ = (
        "sample",
        "args",
        "tasks",
        "cores_per_task",
        "gpus_per_task",
        "timeout",
        "application",
        "app_is_batch_script",
        "_steps",
    )

    def __init__(  # pylint: disable=too-many-arguments,super-init-not-called
        self,
        sample=None,
        args=None,
        tasks=1,
        cores_per_task=1,
        gpus_per_task=0,
        timeout=0,
    ):
        self.sample = sample
        self.args = args
        self.tasks = tasks
        self.cores_per_task = cores_per_task
        self.gpus_per_task = gpus_per_task
        self.timeout = timeout
        self.application = None  # set by Themis instance
        self.app_is_batch_script = True
        self._steps = None

    def __repr__(self):
        """Return string representation of constructor call for this object"""
        return (
            "{}(sample={}, args={!r}, tasks={}, cores_per_task={}, "
            "gpus_per_task={}, timeout={})"
        ).format(
            type(self).__name__,
            self.sample,
            self.args,
            self.tasks,
            self.cores_per_task,
            self.gpus_per_task,
            self.timeout,
        )

    @property
    def steps(self):
        """Dynamically construct the steps making up this run.

        A horrible hack, this requires that the owning themis.Themis
        set ``self.application`` so that ``args[0]`` can be the path
        to the application when the ``Step`` is constructed.
        """
        if self._steps is not None:
            return self._steps
        arguments = [self.application]
        if self.args is not None:
            arguments.extend(shlex.split(self.args))
        return (
            Step(
                arguments,
                self.tasks,
                self.cores_per_task,
                self.gpus_per_task,
                self.timeout,
                self.app_is_batch_script,
            ),
        )

    @steps.setter
    def steps(self, new_steps):
        self._steps = new_steps


class AugmentedRun(CompositeRun):  # pylint: disable=too-few-public-methods
    """A Run object used by Themis's backend with additional fields."""

    __slots__ = CompositeRun.__slots__ + ("status", "result")

    def __init__(self, *args, **kwargs):
        self.status = kwargs.pop("status", None)
        self.result = kwargs.pop("result", None)
        super(AugmentedRun, self).__init__(*args, **kwargs)


def convert_none(value, none_replacement):
    """If value is None, return the replacement value."""
    if value is None:
        return none_replacement
    return value


def validate_application(application):
    """Validate the application; raise a ValueError if it isn't found.

    If application isn't an absolute or relative path, search for it along
    the PATH environment variable.
    """
    if not os.path.exists(application):
        found_application = which(application)
        if found_application is None:
            raise ValueError(
                "{} is not a valid application: it either does not exist"
                " or is not executable.".format(application)
            )
        return found_application
    if not os.path.isfile(application) or not os.access(application, os.X_OK):
        raise ValueError(
            "{} is not a valid application: it either does not exist"
            " or is not executable.".format(application)
        )
    return os.path.abspath(application)


def which(program):
    """Search for `program` along the PATH environment variable."""
    try:
        # works in Python 3; not in 2
        return shutil.which(program)
    except AttributeError:
        pass
    path_var = os.getenv("PATH")
    for path_entry in path_var.split(os.path.pathsep):
        potential_match = os.path.join(path_entry, program)
        if os.path.isfile(potential_match) and os.access(potential_match, os.X_OK):
            return potential_match
    return None


def existing_file(file_path):
    """Check that `file_path` exists; raise a ValueError if it doesn't."""
    if not os.path.isfile(file_path):
        if os.path.isdir(file_path):
            raise ValueError(
                "{!r} is a directory, but a file is required".format(file_path)
            )
        raise ValueError("{!r} is not an existing file".format(file_path))
    return os.path.abspath(file_path)


def match_path(path_pattern, no_dirs):
    """Return files matching `path_pattern`; raise a ValueError if no matches."""
    if no_dirs:
        path_func = existing_file
    else:
        path_func = os.path.abspath
    pattern_matched = glob.glob(path_pattern)
    if not pattern_matched:
        raise ValueError("{!r} does not match anything.".format(path_pattern))
    return [path_func(match) for match in pattern_matched]


def sequence_of_path_patterns(path_patterns, no_dirs=False):
    """Expand an iterable of unix-style path patterns into a sequence of actual paths.

    :param no_dirs: if `True`, raise an error if any of the matches is a directory

    :raises ValueError: if any of the path patterns match no files.
    """
    if path_patterns is None:
        path_patterns = ()
    elif isinstance(path_patterns, str):
        path_patterns = (path_patterns,)
    expanded_paths = set()
    for path in path_patterns:
        for match in match_path(path, no_dirs):
            expanded_paths.add(match)
    return tuple(expanded_paths)


def validate_run_files(input_decks, run_copy, run_symlink):
    """Validate the 3 sets of files/paths."""
    input_decks = set(sequence_of_path_patterns(input_decks, True))
    run_copy = set(sequence_of_path_patterns(run_copy))
    run_symlink = set(sequence_of_path_patterns(run_symlink))
    remove_overlap(
        input_decks, "run_parse", (run_copy, "run_copy"), (run_symlink, "run_symlink")
    )
    remove_overlap(run_copy, "run_copy", (run_symlink, "run_symlink"))
    return (list(input_decks), list(run_copy), list(run_symlink))


def remove_overlap(high_priority_set, high_priority_name, *lower_priority_tuples):
    """Remove the overlap (intersection) between two or more sets."""
    for low_priority_set, low_priority_name in lower_priority_tuples:
        if high_priority_set & low_priority_set:
            warnings.warn(
                "\n{high_priority} and {low_priority} should be distinct, but:\n"
                "{overlap}\nare common to both. "
                "Giving priority to {high_priority}...".format(
                    high_priority=high_priority_name,
                    low_priority=low_priority_name,
                    overlap="\n".join(high_priority_set & low_priority_set),
                ),
                stacklevel=3,
            )
        low_priority_set -= high_priority_set


def type_check(value, *types_to_check):
    """Raise a ValueError if `value` isn't one of `types_to_check`"""
    if not isinstance(value, types_to_check):
        raise TypeError(
            "Expected a type in {}, got {}".format(types_to_check, type(value))
        )
    return value


def range_check(value, min_val=None, max_val=None, name="value"):
    """Check that a variable falls within a range, then return it.

    :param min_val: the minimum acceptable value, or None for no minimum.
    :param min_val: the maximum acceptable value, or None for no maximum.
    :param name: the name of the variable to check. Used for error messages.
    """
    if min_val is not None and value < min_val:
        raise ValueError("{} should be at least {}".format(name, min_val))
    if max_val is not None and value > max_val:
        raise ValueError("{} should be at most {}".format(name, max_val))
    return value


class DirectoryManager(object):
    """Context manager for changing the working directory."""

    def __init__(self, new_directory):
        """Constructor. Saves the name of the directory to change to."""
        self.new_directory = new_directory
        self.old_directory = os.getcwd()

    def __enter__(self):
        """Change to the new working directory."""
        os.chdir(self.new_directory)
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """Return to the original working directory.

        Return False, so the interpreter will re-raise any exceptions
        in the managed block.
        """
        os.chdir(self.old_directory)
        return False


def print_progress_bar(
    iteration, total, suffix="", decimals=1, length=100,
):
    """
    Create a terminal progress bar

    :param iteration: current iteration (int)
    :param total: total iterations (int)
    :param suffix: suffix for the progress bar
    :param decimals: positive number of decimals in percent complete (int)
    :param length: character length of bar (int)
    """
    percent = ("{0:." + str(decimals) + "f}").format(100 * (iteration / float(total)))
    filled_length = int(length * iteration // total)
    for fill_char in ("█", "#"):  # use an ascii character for ascii-only consoles
        progress_bar = fill_char * filled_length + "-" * (length - filled_length)
        try:
            print(
                "|{}| {}% {} ({}/{})".format(
                    progress_bar, percent, suffix, iteration, total
                )
            )
        except UnicodeEncodeError:
            pass
        else:
            break


def import_app_interface(app_interface_path):
    """Import the user's application interface module and return it.

    Don't suppress any exceptions raised by the import.
    """
    if app_interface_path is not None:
        module_name = (
            "app_interface_" + os.path.splitext(os.path.basename(app_interface_path))[0]
        )
        try:
            import importlib.util
        except ImportError:
            import imp

            app_interface = imp.load_source(module_name, app_interface_path)
        else:
            module_spec = importlib.util.spec_from_file_location(
                module_name, app_interface_path
            )
            app_interface = importlib.util.module_from_spec(module_spec)
            module_spec.loader.exec_module(app_interface)
    else:
        app_interface = None
    return app_interface


def get_run_dir(run_id, run_dir_names, sample):
    """Return the run directory for the run given by run_id.

    The default run directory is generated by grouping runs into slots of 10,000
        and then naming the run directory by its run_id.

    :param run_dir_names: A format string that, when formatted by sample, should
        yield the run directory, or None to use the default run directory.
    :param setup_dir: a directory path, which the run directory is relative to
    :param sample: the sample used to format run_dir_names, if run_dir_names is not None
    """
    run_min = (run_id // 10000) * 10000
    run_max = ((run_id // 10000) + 1) * 10000 - 1
    return run_dir_names.format(
        run_id=run_id, run_id_min=run_min, run_id_max=run_max, **sample
    )


def _validate_run_dirs(samples, run_dir_names):
    """Check that the samples will produce unique run directories."""
    set_of_run_dir_names = set()
    for sham_run_id, sample in enumerate(samples):
        try:
            set_of_run_dir_names.add(get_run_dir(sham_run_id, run_dir_names, sample))
        except KeyError:
            raise ValueError(
                "The samples do not have keys matching the run_dir_names format string"
            )
    if len(samples) > len(set_of_run_dir_names):
        raise ValueError(
            "That combination of samples and run_dir_names will "
            "not produce unique run directories."
        )


def validate_runs(runs, run_dir_names=None, application=None, batch_script=True):
    """Validate the type and value of runs and return standardized form."""
    run_dir_names = DEFAULT_RUN_DIR_NAMES if run_dir_names is None else run_dir_names
    runs = list(runs) if not isinstance(runs, Sequence) else runs
    for run in runs:
        # add missing data to the outdated ``Run`` class
        if isinstance(run, Run):
            run.application = application
            run.app_is_batch_script = batch_script
    # validate a reasonable number of run dirs
    _validate_run_dirs([run.sample for run in runs[:5000]], run_dir_names)
    return runs


def validate_samples(samples, run_dir_names=None, check_nonempty=False):
    """Validate the type and value of samples and return standardized form."""
    run_dir_names = DEFAULT_RUN_DIR_NAMES if run_dir_names is None else run_dir_names
    samples = list(samples) if not isinstance(samples, Sequence) else samples
    if not samples and check_nonempty:
        raise ValueError("There are no samples")
    _validate_run_dirs(samples[:5000], run_dir_names)
    return samples


def get_application(application, run_dir, app_is_batch_script):
    """Get the path to the application to the application to execute.

    Usually nothing needs to be done. However, if the application is
    a batch script, it was parsed and copied into the run directory.
    """
    if app_is_batch_script:
        return os.path.join(run_dir, os.path.basename(application))
    return application


class CleanDirectory(object):
    """Context manager for changing the working directory to a new, clean one."""

    def __init__(self, new_directory):
        """Constructor. Saves the name of the directory to change to."""
        self.new_directory = new_directory
        self.old_directory = os.getcwd()

    def go_to_new(self):
        """Create or clear the new directory, and change into it."""
        if os.path.exists(self.new_directory):
            shutil.rmtree(self.new_directory, ignore_errors=True)
        if not os.path.exists(self.new_directory):
            os.mkdir(self.new_directory)
        os.chdir(self.new_directory)
        return self.old_directory

    def go_to_old(self):
        """Return to the old working directory."""
        os.chdir(self.old_directory)

    def __enter__(self):
        """Change to the new working directory and empty it.

        Return the path of the original working directory.
        """
        return self.go_to_new()

    def __exit__(self, exc_type, exc_value, traceback):
        """Return to the original working directory.

        Return False, so the interpreter will re-raise any exceptions
        in the managed block.
        """
        self.go_to_old()
        return False


def clean_directory_decorator(new_directory=None):
    """Wrap a function in a CleanDirectory context.

    If the name of a new directory is not given,
    use the name of the passed function.
    """

    def decorator(func):
        dir_context = new_directory or func.__name__

        @functools.wraps(func)
        def nested(*args, **kwargs):
            with CleanDirectory(dir_context):
                return func(*args, **kwargs)

        return nested

    return decorator


def _alarm_exception(signum, frame):
    """Signal handler function."""
    raise OSError("Alarm handler called")


def timeout_decorator(timeout):
    """Raise a TimeoutError if a function takes more than `timeout` seconds."""

    def decorator(func):
        """Wrap a function in calls to SIGALRM."""

        @functools.wraps(func)
        def nested(*args, **kwargs):
            """Request an alarm, call a function, disable the alarm."""
            signal.signal(signal.SIGALRM, _alarm_exception)
            signal.alarm(timeout)
            try:
                return_val = func(*args, **kwargs)
            finally:
                signal.alarm(0)  # disable the alarm, the function finished
            return return_val

        return nested

    return decorator


def add_parameterfile_arg(parser, nargs=None):
    """Add a parameterfile argument to an `argparse.ArgumentParser`."""
    parser.add_argument(
        "parameterfile",
        type=str,
        help=(
            "Path to the file defining the parameters for each run."
            " The file is assumed to be a CSV with a header row."
        ),
        nargs=nargs,
    )


def add_common_creation_args(parser):
    """Add argparse.ArgumentParser arguments common to __main__.py and laf.py"""
    parser.add_argument(
        "-r",
        "--run-dir-names",
        type=str,
        help=(
            "The Python format string yielding the run directories. "
            "Default behavior is to put each run into `runs/####`."
        ),
        metavar="FORMATSTRING",
    )
    parser.add_argument(
        "--parse",
        nargs="+",
        type=str,
        help=(
            "Paths to text files to parse and copy into each run directory. "
            "The files are parsed for variables declared with '%%%%'. "
            "For example, '%%%%my_sample_1%%%%' will be replaced with the value "
            "of 'my_sample_1' for each run."
        ),
        metavar="FILE",
        dest="run_parse",
    )
    parser.add_argument(
        "--symlink",
        nargs="+",
        type=str,
        help=(
            "Paths to files/directories to symlink into each run directory. "
            "E.g. `--symlink *` will symlink everything in the current working "
            "directory. Default behavior is not to symlink anything. "
        ),
        metavar="PATH",
        dest="run_symlink",
    )
    parser.add_argument(
        "--copy",
        nargs="+",
        type=str,
        help=(
            "Paths to files/directories to hard-copy into each run directory. "
            "E.g. `--copy *` will copy everything in the current working directory. "
            "Default behavior is not to copy anything. "
        ),
        metavar="PATH",
        dest="run_copy",
    )


def read_csv(path):
    """Read a CSV and return an iterable of dictionaries representing the rows."""
    with open(path, **csvargs) as csvfile:
        reader = csv.DictReader(csvfile)
        for row in reader:
            yield row



"""
This module defines the EnsembleDatabase class and supporting functions.

The EnsembleDatabase class is used to connect and interact with
the ensemble's backend. This backend is used all over this package--
by both workers and by the user-facing API.

The backend (whatever it is) is used both as a simple store of information
and as concurrency control. Most of the operations `must` be atomic.

The reference EnsembleDatabase implementation is the SQLiteDatabase class,
defined in this module.
"""

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import pickle

from themis.database.utils import (
    EnsembleDatabase,
    SQLiteMixin,
    cursorcloser,
    sql_transaction,
    unpickle_or_none,
)
from themis.utils import CompositeRun, AugmentedRun


class SqlRunInfo(object):  # pylint: disable=too-few-public-methods
    """Represents and stores the information needed by workers to execute a run."""

    __slots__ = ("_row", "run_id", "restarts", "status", "steps", "sample")

    def __init__(self, row):
        self._row = row
        self.run_id = row[0]
        self.restarts = row[1]
        self.status = row[2]
        self.steps = pickle.loads(row[3])
        self.sample = pickle.loads(row[4])

    def __repr__(self):
        """Return str representation of constructor call"""
        return "{}({})".format(type(self).__name__, self._row)

    @property
    def run(self):
        """Return a ``CompositeRun`` object."""
        return CompositeRun(self.sample, self.steps)


class SQLiteDatabase(SQLiteMixin, EnsembleDatabase):
    """SQLite backend for the ensemble, using the sqlite3 standard library module.

    This is the reference EnsembleDatabase implementation.

    The database is stored as a local file. Only one connection can execute a
    write command at a time. See the SQLite documentation for more information.

    Note that SQLite connections cannot be safely be shared between threads.
    For multithreaded environments, each thread should have its own copy
    of an instance; multiple threads should not have references to the same instance.

    This database has one table, RunInfo. This table holds all the information about
    the runs in the ensemble. Each row is one run. The columns are as follows:
        runID: a unique nonnegative integer identifying the run. The primary key of the
            table. New rows have their ID generated automatically by SQLite.
        restarts: the number of times the run has been restarted due to a failure.
        step_num: the current step that the run is on. Each run consists of one or more
            steps.
        taken_bool: a boolean indicating whether the run is available to be grabbed
            for execution. 1 if the run is unavailable, 0 else.
        completion_status: an integer indicating the status of the run.
            One of the RUN_* enums.
        arg: the arguments for the run
        tasks: the number of tasks for the run
        cores_per_task: the cores per task for the run
        gpus_per_task: the gpus per task for the run
        timeout: the timeout for the run
        sample: the sample for the run, pickled
        result: the result for the run, pickled


    :param db_path: the path to the database. If the database does not exist, it
        will be created.
    """

    def __init__(self, db_path):
        """Constructor. Opens a connection to the database.

        The connection is only closed on object destruction.
        """
        super(SQLiteDatabase, self).__init__(db_path, timeout=200)
        self._table_names.append("RunInfo")

    def create(self):
        """Create and populate the RunInfo table.

        The RunInfo table makes up the bulk of the database; it stores all of
        the run-specific information. That includes the status of the current run,
        its command-line arguments, its resources, sample, and result, and more.

        Most of these are stored as integers. However, the samples and results
        are stored in pickled form as the BLOB type.
        """
        create_query = (
            "CREATE TABLE RunInfo (runID INTEGER PRIMARY KEY, "
            "restarts SMALLINT DEFAULT 0, step_num SMALLINT DEFAULT 0, "
            "taken_bool SMALLINT DEFAULT 0, "
            "completion_status SMALLINT DEFAULT {}, "
            "sample BLOB, steps BLOB, result BLOB)"
        ).format(int(self.RUN_QUEUED))
        with cursorcloser(self._db_handle) as cursor:
            with sql_transaction(cursor):
                cursor.execute(create_query)
                # create indices to improve performance on big DBs
                cursor.execute(
                    "CREATE INDEX new_run_index ON "
                    "RunInfo(taken_bool, completion_status)"
                )
                cursor.execute(
                    "CREATE INDEX completion_index ON RunInfo(completion_status)"
                )

    def add_result(self, run_ids, results):
        """Add a result or sequence of results to the database.

        :param run_ids: an int run_id or iterable of int run_ids,
            indicating which runs the results belong to
        :param results: a single result or iterable of results,
            matching in this respect (and in the length of the iterable)
            with run_ids
        """
        if isinstance(run_ids, int):
            run_ids = (run_ids,)
            results = (results,)
        results = [
            pickle.dumps(result, protocol=pickle.HIGHEST_PROTOCOL) for result in results
        ]
        add_result_query = "UPDATE RunInfo SET result = ? WHERE runID = ?"
        with cursorcloser(self._db_handle) as cursor, sql_transaction(cursor):
            cursor.executemany(add_result_query, zip(results, run_ids))

    def get_result(self, run_ids):
        """Return the result values for a run or sequence of runs.

        :param run_ids: the run_ids to fetch results for
        :return: a mapping from run_ids to results
        """
        if isinstance(run_ids, int):
            run_ids = (run_ids,)
        pickled_results = []
        with cursorcloser(self._db_handle) as cursor, sql_transaction(cursor):
            for run_id in run_ids:
                cursor.execute("SELECT result FROM RunInfo WHERE runID = ?", (run_id,))
                pickled_results.append(cursor.fetchone())
        return {
            run_id: pickle.loads(result[0])
            if result is not None and result[0] is not None
            else None
            for run_id, result in zip(run_ids, pickled_results)
        }

    def _fetch_results(self, include_none):
        query = "SELECT runID, sample, steps, result, completion_status FROM RunInfo"
        if not include_none:
            query += " WHERE result IS NOT NULL"
        with cursorcloser(self._db_handle) as cursor, sql_transaction(cursor):
            cursor.execute(query)
            return cursor.fetchall()

    def as_mappings(self, include_none):
        """Return info about all runs as a list of mappings"""
        result_rows = self._fetch_results(include_none)
        return [
            {
                "run_id": row[0],
                "sample": pickle.loads(row[1]),
                "steps": pickle.loads(row[2]),
                "result": unpickle_or_none(row[3]),
                "status": row[4],
            }
            for row in result_rows
        ]

    def add_runs(self, runs):
        """Add an iterable of new sample points to the database.

        The run IDs of the new points is determined when they are added.
        """
        with cursorcloser(self._db_handle) as cursor, sql_transaction(cursor):
            cursor.executemany(
                "INSERT INTO RunInfo(sample, steps) VALUES (?, ?)",
                (
                    (
                        pickle.dumps(run.sample, protocol=pickle.HIGHEST_PROTOCOL),
                        pickle.dumps(run.steps, protocol=pickle.HIGHEST_PROTOCOL),
                    )
                    for run in runs
                ),
            )

    def get_run_info(self, run_id):
        """Return the RunInfo object for the run specified by run_id."""
        with cursorcloser(self._db_handle) as cursor, sql_transaction(cursor):
            cursor.execute(
                "SELECT runID, restarts, step_num, steps, "
                "sample FROM RunInfo WHERE runID = ?",
                (run_id,),
            )
            result = cursor.fetchone()
        if result is None:
            return None
        return SqlRunInfo(result)

    def user_facing_run_info(self, run_ids):
        """Return a mapping from run IDs to AugmentedRun objects for `run_ids`."""
        rows = []
        with cursorcloser(self._db_handle) as cursor, sql_transaction(cursor):
            for run_id in run_ids:
                cursor.execute(
                    "SELECT runID, completion_status, result, sample, "
                    "steps FROM RunInfo WHERE runID = ?",
                    (run_id,),
                )
                row = cursor.fetchone()
                if row is not None:
                    rows.append(row)
        return {
            row[0]: AugmentedRun(
                sample=pickle.loads(row[3]),
                status=row[1],
                result=pickle.loads(row[2]) if row[2] is not None else None,
                steps=pickle.loads(row[4]),
            )
            for row in rows
        }

    def set_run_status(self, run_ids, status_tuples):
        """Update the statuses of runs.

        The length of run_ids and status_tuples should be the same.

        :param run_ids: an iterable of int run_ids
        :param status_tuples: an iterable of 4-tuples with the following format:
            (step, completion code, active bool, number of restarts)
        """
        if not run_ids:
            return
        with cursorcloser(self._db_handle) as cursor, sql_transaction(cursor):
            cursor.executemany(
                "UPDATE RunInfo SET step_num = ?, completion_status = ?, "
                "restarts = ? WHERE runID = ?",
                (
                    status_tup + (run_id,)
                    for run_id, status_tup in zip(run_ids, status_tuples)
                ),
            )

    def disable_runs(self, run_ids):
        """Mark the runs indicated by run_ids as disabled.

        Disabled runs can never be collected by the new_runs method,
        unless they are later re-enabled by a call to enable_runs.
        """
        self._set_run_taken_bool(run_ids, 1)

    def enable_runs(self, run_ids):
        """Mark the runs indicated by run_ids as enabled.

        Unless the runs were disabled, this method has no effect.
        """
        self._set_run_taken_bool(run_ids, 0)

    def _set_run_taken_bool(self, run_ids, new_status):
        """Set the taken_bool of run_ids to new_status."""
        with cursorcloser(self._db_handle) as cursor, sql_transaction(cursor):
            cursor.executemany(
                "UPDATE RunInfo SET taken_bool = ? WHERE runID = ?",
                ((new_status, run_id) for run_id in run_ids),
            )

    def enable_all(self):
        """Mark all runs as enabled."""
        with cursorcloser(self._db_handle) as cursor, sql_transaction(cursor):
            cursor.execute(
                "UPDATE RunInfo SET taken_bool = 0 WHERE "
                "taken_bool = 1 AND completion_status = ?",
                (self.RUN_QUEUED,),
            )

    def new_runs(self, limit):
        """Return a sequence (with length <= limit) of new RunInfo objects.

        These RunInfo objects are collected atomically, so that no other
        processes or threads calling this method will get the same objects.
        """
        if limit == 0:
            return ()
        with cursorcloser(self._db_handle) as cursor, sql_transaction(
            cursor, immediate=True
        ):
            # need to select runIDs and update those runs'
            # taken_bool in a single transaction
            # otherwise another DB connection could take them
            cursor.execute(
                "SELECT runID, restarts, step_num, steps, sample FROM RunInfo "
                "WHERE taken_bool == 0 AND completion_status <= ? LIMIT ?",
                (self.RUN_QUEUED, limit),
            )
            results = cursor.fetchall()
            # now change the runs with intermediate status to their final status
            cursor.executemany(
                "UPDATE RunInfo SET taken_bool = 1 WHERE runID = ?",
                [(row[0],) for row in results],
            )
        return tuple(SqlRunInfo(result) for result in results)

    def _runs_with_status(self, statuses, count=False):
        """Return a tuple containing all the run ID's of runs that have a
        certain completion status.

        Alternatively, if `count` is True, return the number of runs with one of the
        given statuses.

        :param statuses: an iterable of EnsembleDatabase.RUN_* status objects.
        """
        select_run_id_status = "SELECT {} FROM RunInfo"
        if count:
            select_run_id_status = select_run_id_status.format("COUNT(runID)")
        else:
            select_run_id_status = select_run_id_status.format("runID")
        if statuses:
            select_run_id_status += (
                " WHERE completion_status IN (" + ("?, " * (len(statuses) - 1)) + "?)"
            )
        with cursorcloser(self._db_handle) as cursor, sql_transaction(cursor):
            cursor.execute(select_run_id_status, statuses)
            result = cursor.fetchall()
        if count:
            return result[0][0]
        # the run_ids are returned as 1-tuples, so we must unpack them
        return tuple(one_tuple[0] for one_tuple in result)

    def runs_with_completion_status(self, *statuses):
        """Return an iterable of run ids with one of the given completion statuses.

        :param statuses: zero or more EnsembleDatabase.RUN_* status objects.
            If status is empty, return all run ID's. Otherwise, return any run ID
            with a status contained in the statuses parameter.
        """
        return self._runs_with_status(statuses)

    def count_runs_by_completion(self, *statuses):
        """Return the number of runs with one of the given completion statuses.

        :param statuses: same meaning as in runs_with_completion_status.
        """
        return self._runs_with_status(statuses, True)

    def mark_runs_to_restart(self, run_ids, hard):
        """Mark the runs indicated by run_ids as eligible for restart."""
        if hard:
            set_step_num = "step_num = 0, "
        else:
            set_step_num = ""
        params = [(self.RUN_QUEUED, run_id, self.RUN_QUEUED) for run_id in run_ids]
        with cursorcloser(self._db_handle) as cursor, sql_transaction(cursor):
            cursor.executemany(
                "UPDATE RunInfo SET restarts = 0, taken_bool = 0, "
                + set_step_num
                + "completion_status = ? WHERE runID = ? AND "
                "completion_status > ?",
                params,
            )

    def mark_runs_to_kill(self, run_ids):
        """Mark that the runs indicated by run_ids should be killed."""
        with cursorcloser(self._db_handle) as cursor, sql_transaction(cursor):
            cursor.executemany(
                "UPDATE RunInfo SET completion_status = ?, taken_bool = 1 "
                "WHERE runID = ? AND completion_status <= ?",
                zip(
                    (self.RUN_KILLED for _ in run_ids),
                    run_ids,
                    (self.RUN_QUEUED for _ in run_ids),
                ),
            )

    def runs_to_kill(self):
        """Return a sequence of run IDs that should be killed."""
        return self.runs_with_completion_status(self.RUN_KILLED)



"""
This script prepares an ensemble for execution, and launches the
individual applications making up the ensemble.

Many variable names are used repeatedly throughout this module. They are
intended to have a consistent meaning. Some of them are as follows:
    :param ensemble_db: an EnsembleDatabase object
    :param app_spec: a constant mapping containing information about the ensemble.
        Should have a fixed set of keys; see the ensemble.database module for more.
    :param app_interface: a user's application interface module, or None.
"""
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import os
import sys
import time
import logging
import socket
import argparse
import contextlib

if __name__ == "__main__":
    # when running as a script, sys.path needs to be set for imports
    sys.path.insert(0, os.path.abspath(__file__).rsplit(os.sep, 4)[0])

# pylint: disable=wrong-import-position
import themis.utils
from themis import database
from themis.versions import reprlib
from themis import backend
from themis.backend.network import BackendClient, CERT_NAME
from themis.backend.worker import schedulers
from themis.backend.worker import network


LOGGER = logging.getLogger(__name__)


class ProgressSaver(object):
    """Instances of this class are used to save progress in the submission loop.

    The interface consists only of the __call__ method, which should be
    called regularly. It only saves progress if progress hasn't been in the past
    ``frequency`` seconds; otherwise, does nothing.

    :param max_failed_runs: The maximum number of tolerated failures
    :param frequency: the frequency (in seconds) of saves
    """

    def __init__(self, server, scheduler, max_failed_runs, frequency):
        """Initialize a new instance."""
        self._server = server
        self._scheduler = scheduler
        self._max_failed_runs = max_failed_runs
        self._save_interval = frequency
        self._time_of_last_save = time.time()

    def __repr__(self):
        """Return str representation of constructor call"""
        return "{}({}, {}, {}, {})".format(
            type(self).__name__,
            self._server,
            self._scheduler,
            self._max_failed_runs,
            self._save_interval,
        )

    def __call__(self, ensemble_db, status_store):
        """Save server and scheduler progress; return True if an exit condition
        was reached.

        Only saves progress if progress hasn't been in the past ``frequency``
        seconds; otherwise, does nothing.

        The three exit conditions recognized:
        1)  The server or scheduler thread dies
        2)  A STOP_ENSEMBLE message is received
        3)  Too many runs fail

        :param curr_time: the current Unix standard time.
        """
        curr_time = time.time()
        if curr_time - self._time_of_last_save > self._save_interval:
            self._time_of_last_save = curr_time
            return self.save(ensemble_db, status_store)
        return False

    def save(self, ensemble_db, status_store):
        """Save progress."""
        if self._check_alive() or self._check_messages(status_store, ensemble_db):
            return True
        # get newly failed runs from the scheduler. The set enforces uniqueness
        completed_ids_and_status = self._scheduler.save_progress(ensemble_db)
        num_new_failed_runs = sum(
            [
                status == ensemble_db.RUN_FAILURE
                for _, status in completed_ids_and_status
            ]
        )
        self._server.remove_completed_runs(
            [run_id for run_id, _ in completed_ids_and_status]
        )
        self._server.save_progress(ensemble_db)
        return self._check_failed_runs(status_store, num_new_failed_runs)

    def _check_alive(self):
        """Return True if the server or scheduler has died."""
        if not (self._server.is_alive() and self._scheduler.is_alive()):
            LOGGER.error("Threading exception, exiting...")
            return True
        return False

    def _check_messages(self, status_store, ensemble_db):
        """Check user messages for killing runs or stopping the ensemble.

        Kill any runs required by the messages.

        If a STOP_ENSEMBLE message is received, return True once current runs
        have cleared (i.e. the ensemble has stopped).
        """
        runs_to_kill = ensemble_db.runs_to_kill()
        LOGGER.debug("Killing runs %s", reprlib.repr(runs_to_kill))
        self._scheduler.kill_runs(runs_to_kill)
        if status_store.fetch()["stop"]:
            # if a 'complete_existing_runs' message is received, shut down
            LOGGER.info("Stop ensemble message received")
            self._scheduler.complete_existing_runs()
            return True
        return False

    def _check_failed_runs(self, status_store, num_new_failed_runs):
        """Return True if the number of failed runs exceeds the maximum allowed."""
        if self._max_failed_runs > -1:
            total_failed_runs = status_store.increment_failed_runs(num_new_failed_runs)
            if total_failed_runs > self._max_failed_runs:
                # pass None for app interface so post_ensemble isn't called
                LOGGER.error(
                    "Total failed runs (%s) exceeds maximum of %s. Exiting...",
                    total_failed_runs,
                    self._max_failed_runs,
                )
                # don't call post_ensemble--just let existing runs complete
                self._scheduler.complete_existing_runs()
                return True
        return False


# pylint: disable=too-many-arguments
def enter_submission_loop(
    ensemble_db, status_store, app_spec, queue_size, do_prep, do_post
):
    """Enter a loop of submitting runs and checking for new runs.

    Aside from errors, there are a handful of exit conditions:
    1)  There are no new runs, all existing runs have completed,
            AND the user's post_ensemble has been called and did
            not return any new runs.
    2)  A user called for a STOP_ENSEMBLE; in this case, the loop exits
            once all existing runs have completed.
    3)  The maximum number of failed runs has been exceeded.

    :return: an iterable, with entries being either
        Profiler objects from the cProfile module, or None.
    """
    # create a Scheduler object which takes in run IDs and runs them
    with backend.managed_shutdown(
        network.WorkerServer(CERT_NAME, app_spec), ensemble_db
    ) as server, backend.managed_shutdown(
        schedulers.setup_scheduler(
            app_spec, server.server_address, queue_size, do_prep, do_post
        ),
        ensemble_db,
    ) as scheduler, backend.RunFeeder(
        # Pick a reasonable size for the cache of new runs
        ensemble_db,
        int(2.5 * queue_size),
    ) as run_feeder:
        LOGGER.info("Beginning to submit runs.")
        LOGGER.info("Server listening on %s", server.server_address)
        progress_saver = ProgressSaver(
            server, scheduler, app_spec["max_failed_runs"], app_spec["save_interval"]
        )
        # fetch new runs from the RunFeeder whenever the scheduler has less than
        # min_pending_runs (some reasonable factor of max_concurrency) to execute
        min_pending_runs = max(2, queue_size)
        while True:
            pending_runs = scheduler.count_pending_runs()
            if pending_runs < min_pending_runs:
                # get a sequence of new RunInfo objects to submit, and their statuses
                new_runs = run_feeder.new_runs(min_pending_runs - pending_runs)
                # submit the runs (whether new or restarts) to the scheduler and server
                scheduler.submit(new_runs)
                server.submit(new_runs)
                run_feeder.poll_cache()
                if run_feeder.is_empty() and scheduler.done():
                    # there are no new runs to submit
                    if not run_feeder.fill_cache():
                        # all existing runs have completed, and no new runs yet
                        # Now check if the user's post_ensemble adds new runs
                        LOGGER.info("Ensemble complete. Shutting down and exiting...")
                        break
            # check for messages and save progress
            if progress_saver(ensemble_db, status_store):
                # the call only returns true if exit condition #2 or #3 occurs
                LOGGER.info("Stop condition triggered, exiting submission loop")
                break
            time.sleep(0.02)
        return (scheduler.profiler, server.profiler)


def main(app_spec, ensemble_db, queue_size):
    """Perform all setup actions before entering the main event loop.

    :param max_runtime: the max max_runtime (in minutes) for this submitter
    """
    status_store = database.get_status_store(app_spec["setup_dir"])
    LOGGER.info("Using database: %s", ensemble_db)
    # import the application interface
    app_interface = themis.utils.import_app_interface(app_spec["app_interface"])
    do_post = hasattr(app_interface, "post_run")
    do_prep = hasattr(app_interface, "prep_run") or len(app_spec["run_parse"]) > 2
    # enter a loop of submitting runs and then checking for new runs
    return enter_submission_loop(
        ensemble_db, status_store, app_spec, queue_size, do_prep, do_post
    )


def setup_parser():
    """Set up the command-line parser."""
    parser = argparse.ArgumentParser(
        description=(
            "Execute runs. Requires a database "
            "holding information about the ensemble."
        )
    )
    parser.add_argument(
        "--id",
        metavar="N",
        type=str,
        help="the (unique) id of this scheduler run",
        default="0",
    )
    parser.add_argument(
        "-c",
        "--max-concurrency",
        metavar="N",
        type=int,
        help="the maximum number of runs to execute concurrently",
        required=True,
    )
    parser.add_argument(
        "--server-args", nargs="*", help="server connection information", type=str
    )
    parser.add_argument(
        "--setup-dir", help="Path to the Themis setup directory", default=os.curdir,
    )
    return parser


def get_identifier(args):
    """Return the ID of this instance and the number of cores and GPUs it controls."""
    if os.getenv("FLUX_TREE_ID") is not None:
        # if running inside of Flux tree, there will be no resource args,
        # only env vars exported by Flux
        identifier = os.getenv("FLUX_TREE_ID")
        sys.stderr = open("themis_worker_{}.log".format(identifier), "a")
        sys.stdout = sys.stderr
    else:
        identifier = str(args.id)
    return identifier


@contextlib.contextmanager
def exit_on_signal():
    """On SignalCaught exception, exit Python."""
    try:
        yield
    except backend.SignalCaught as sig:
        LOGGER.info("Signal %i received, exiting...", sig.signum)
        raise


def setup():
    """Collect this submitter's setup and pass it to main().

    Profile main if the profiling env var is set.
    """
    args = setup_parser().parse_args()
    identifier = get_identifier(args)
    LOGGER.info("Worker starting...")
    LOGGER.info("Received args %s", args)
    if "FLUX_URI" in os.environ:
        LOGGER.info(
            "Flux remote URI is %s, local URI is %s",
            "ssh://"
            + socket.gethostname()
            + os.environ["FLUX_URI"].replace("local://", ""),
            os.environ["FLUX_URI"],
        )
    app_spec = database.get_app_spec(args.setup_dir)
    backend.handle_signals()
    with exit_on_signal():
        LOGGER.info("Maximum concurrent runs: %i", args.max_concurrency)
        if args.server_args:
            ensemble_db = BackendClient(CERT_NAME, *args.server_args)
        else:
            ensemble_db = database.default_database(app_spec)
        profiler, additional_profilers = backend.profile_from_new(
            main, app_spec, ensemble_db, args.max_concurrency
        )
        backend.dump_profile_data(
            "submitter_profile_stats_{}.pstat".format(identifier),
            profiler,
            *additional_profilers
        )
    LOGGER.info("Exited cleanly.")


if __name__ == "__main__":
    with backend.managed_logging(sys.stderr):
        setup()



"""
This script prepares an individual application run for execution.

That consists of copying, symlinking, and parsing files, then
calling the user's prep_run function.
"""
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import os
import shutil
import re
import sys
import logging

try:
    import jinja2
except ImportError:
    JINJA2_AVAIL = False
else:
    JINJA2_AVAIL = True

if __name__ == "__main__":
    # when running as a script, sys.path needs to be set for imports
    sys.path.insert(0, os.path.abspath(__file__).rsplit(os.sep, 4)[0])

# pylint: disable=wrong-import-position
from themis.utils import DirectoryManager
from themis.backend.worker import utils
from themis import resource


def user_prep_run(run_dir, prep_run):
    """Call the user's prep_run function, if it exists, and return its return value."""
    if callable(prep_run):
        with DirectoryManager(run_dir):
            prep_run()


def remove(path):
    """Remove a file or directory at `path`"""
    if os.path.isfile(path):
        os.remove(path)
    elif os.path.isdir(path):
        shutil.rmtree(path)


def populate_run_dir(symlinks, copies, input_decks, sample, run_dir):
    """Create the run directory and populate it.

    The directory is populated with symlinks of every file/dir in `symlinks`,
    hard copies of every file/dir in `copies`,
    and _parsed_ hard copies of every file in input_decks.
    """
    # symlink required files into the new directory
    for source in symlinks:
        sym_target_path = os.path.join(run_dir, os.path.basename(source))
        # if a file exists with the same name as the new symlink, delete it
        if os.path.lexists(sym_target_path):
            os.remove(sym_target_path)
        os.symlink(source, sym_target_path)
    for source in copies:
        copy_dest = os.path.join(run_dir, os.path.basename(source))
        if os.path.isdir(source):
            if os.path.exists(copy_dest):
                remove(copy_dest)
            shutil.copytree(source, copy_dest, symlinks=True)
        else:
            shutil.copy(source, copy_dest)
    # hard copy input deck files into the new directory
    for source in input_decks:
        dest = os.path.join(run_dir, os.path.basename(source))
        # if copy_dest (the target of the new input deck) exists, overwrite it
        parse_input_deck(source, dest, sample)


def parse_input_deck(template, output_path, sample):
    """Check for jinja availability and forward on to the right function."""
    if JINJA2_AVAIL:
        parse_input_deck_jinja(template, output_path, sample)
    else:
        parse_input_deck_base(template, output_path, sample)
    shutil.copystat(template, output_path)


def _alternate_input_deck_syntax(line, sample):
    """Limited upport for old-style "UQP_VARIABLE" parsing.

    Takes a line like "var = ... # UQP_VARIABLE = foo", converts
    it to "var = %%foo%%", and return the modified line.
    """
    if "UQP_VARIABLE" not in line:
        return line
    pattern_replacements = [
        (r"=.*#\s*UQP_VARIABLE\s*=\s*{}".format(key), "= %%{}%%".format(key))
        for key in sample.keys()
    ]
    for pattern, replacement in pattern_replacements:
        line = re.sub(pattern, replacement, line)
    return line


def parse_input_deck_base(template, output_path, sample):
    """Parse a text file for variables declared with '%%' characters.

    Also supports old-style "UQP_VARIABLE" parsing.

    :param template: path to the template input deck to parse
    :param output_path: path to write the parsed template
    :param sample: mapping from variable names to values
    """
    if not sample:
        shutil.copyfile(template, output_path)
        return
    patterns_to_values = {"%%{}%%".format(key): str(val) for key, val in sample.items()}
    sample_regex = re.compile(
        "|".join(["({})".format(key) for key in patterns_to_values.keys()])
    )
    lines = []
    # collect and format lines in the template
    with open(template) as template_fd:
        for line in template_fd:
            line = _alternate_input_deck_syntax(line, sample)
            lines.append(
                sample_regex.sub(lambda match: patterns_to_values[match.group(0)], line)
            )
    # write formatted lines out to the target
    with open(output_path, "w") as output_fd:
        output_fd.writelines(lines)


def parse_input_deck_jinja(template_path, output_path, sample):
    """Parse a file using jinja templating."""
    env = jinja2.Environment(
        loader=jinja2.FileSystemLoader(os.path.dirname(template_path)),
        autoescape=True,
        keep_trailing_newline=True,
        newline_sequence=os.linesep,
        variable_start_string="%%",
        variable_end_string="%%",
    )
    template = env.get_template(os.path.basename(template_path))
    with open(output_path, "w") as file_handle:
        template.stream(sample).dump(file_handle)


def preparation(prep_run, run_dir, app_spec, sample, steps):
    """Complete the preparation for a single run.

    Populate the run directory, and call user's prep_run function.
    """
    resource_mgr = resource.identify_resource_manager(app_spec["resource_mgr"])
    populate_run_dir(
        app_spec["run_symlink"],
        app_spec["run_copy"],
        app_spec["run_parse"],
        sample,
        run_dir,
    )
    for step in steps:
        if step.batch_script:
            step.batch_script = False  # set so build_cmd works
            sample["themis_launch"] = " ".join(resource_mgr.build_cmd(step))
            step.batch_script = True
            parse_input_deck(
                step.args[0],
                os.path.join(run_dir, os.path.basename(step.args[0])),
                sample,
            )

    user_prep_run(run_dir, prep_run)


def main(run_id, server_args, setup_dir):
    """Collect setup information and pass it to the `preparation` function."""
    _, app_interface, run_dir, app_spec, run_info = utils.setup_script(
        run_id, server_args, setup_dir
    )
    prep_run = getattr(app_interface, "prep_run", None)
    logger = logging.getLogger(__name__)
    if not JINJA2_AVAIL:
        logger.warning(
            "jinja2 not available, using built-in support for template files"
        )
    else:
        logger.debug("Using jinja2 for file templates...")
    logger.info("Populating run directory with files and calling user's prep_run")
    preparation(prep_run, run_dir, app_spec, run_info.sample, run_info.steps)
    logger.debug("Run setup complete")


if __name__ == "__main__":
    utils.profile_worker(main, "prepper")



"""This module renames imports for Python 2/3 compatibility."""

import sys

# pylint: skip-file

if sys.version_info.major >= 3:
    basestring = str
    import socketserver
    import pickle
    import reprlib
    import http.server as http_server
    import http.client as http_client
    from collections import UserDict
    from urllib.parse import parse_qs, parse_qsl, urlencode
    try:
        from collections.abc import Mapping, Sequence
    except ImportError:
        from collections import Mapping, Sequence

    csvargs = {"newline": ""}  # arguments for opening files with the `csv` module

else:
    basestring = basestring
    import SocketServer as socketserver
    import cPickle as pickle
    import repr as reprlib
    import BaseHTTPServer as http_server
    import httplib as http_client
    from UserDict import UserDict
    from urlparse import parse_qs, parse_qsl
    from urllib import urlencode
    from collections import Mapping, Sequence

    csvargs = {}  # arguments for opening files with the `csv` module



"""The ``themis.runtime`` module provides functions and operations for
applications launched by Themis.
"""

import os

from themis.backend.network import CERT_NAME
from themis.backend.worker import prepper, network
from themis.utils import (
    URI_ENV_VAR,
    RUNID_ENV_VAR,
    SETUPDIR_ENV_VAR,
    URI_SPLITCHAR,
    EXECDIR_ENV_VAR,
)
import themis


_RUNTIME_ERRMSG = "This function must be called from an application launched by Themis"


def _verify_themis_runtime():
    """Check environment variables and return a (run ID, themis connection) pair."""
    for env_var in (URI_ENV_VAR, RUNID_ENV_VAR, SETUPDIR_ENV_VAR, EXECDIR_ENV_VAR):
        if env_var not in os.environ:
            raise RuntimeError(_RUNTIME_ERRMSG)
    run_id = os.environ[RUNID_ENV_VAR]
    host, _, port = os.environ[URI_ENV_VAR].rpartition(URI_SPLITCHAR)
    cert_path = os.path.join(os.environ[EXECDIR_ENV_VAR], CERT_NAME)
    client = network.WorkerClient(cert_path, host, port)
    return (run_id, client)


def fetch_run_id():
    """Return the integer ID of the current run.

    :raises RuntimeError: if called outside of the Themis runtime environment.
    """
    if RUNID_ENV_VAR not in os.environ:
        raise RuntimeError(_RUNTIME_ERRMSG)
    return int(os.environ[RUNID_ENV_VAR])


def fetch_run():
    """Return the ``themis.CompositeRun`` object for the current run.

    :raises RuntimeError: if called outside of the Themis runtime environment.
    """
    run_id, client = _verify_themis_runtime()
    _, run_info = client.get_worker_setup(run_id)
    return run_info.run


def themis_handle():
    """Return a ``themis.Themis`` handle for the current ensemble.

    :raises RuntimeError: if called outside of the Themis runtime environment.
    """
    if SETUPDIR_ENV_VAR not in os.environ:
        raise RuntimeError(_RUNTIME_ERRMSG)
    return themis.Themis(setup_dir=os.environ[SETUPDIR_ENV_VAR])


def set_result(result):
    """Set the result for the current run.

    ``result`` should be an arbitrary Python object consisting only
    of built-in types, rather than custom classes.

    :raises RuntimeError: if called outside of the Themis runtime environment.
    """
    run_id, client = _verify_themis_runtime()
    client.add_result(run_id, result)


def parse_file(source, destination, sample=None):
    """Parse and token-replace a text file based on the key-value pairs in ``sample``.

    :param source: the path to the text file to parse.
    :param destination: the destination for the resulting parsed file
    :param sample: the key-value pairs to use for the token-replacement operation.
        If ``None``, fetch the sample by calling ``themis.runtime.run().sample``.
    """
    if sample is None:
        sample = fetch_run().sample
    prepper.parse_input_deck(source, destination, sample)



"""Script comprising the entry point to the backend of themis"""

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import os
import sys
import logging
import functools
import argparse
import time
import contextlib
import signal
import socket

if __name__ == "__main__":
    # when running as a script, sys.path needs to be set for imports
    sys.path.append(os.path.abspath(__file__).rsplit(os.sep, 3)[0])


# pylint: disable=wrong-import-position
from themis import database
from themis import utils
from themis import resource
from themis import backend
from themis.backend import network
from themis.backend import RunFeeder, SignalCaught


LOGGER = logging.getLogger(__name__)


def create_new_allocation(status_store, resource_mgr, args):
    """Create a new allocation to follow this one."""
    curr_allocation = status_store.decrement_repeats(args.alloc_id)
    if curr_allocation.repeats == -1:
        LOGGER.info("No remaining allocation repeats. Exiting...")
        return
    LOGGER.info("Replicating existing allocation...")
    backend_commands = [
        resource_mgr.commands_to_launch_backend(
            curr_allocation.timeout,
            args.parallelism,
            args.alloc_id,
            args.early_stop,
            args.multiple,
            args.setup_dir,
            args.max_concurrency,
        )
    ]
    resource_mgr.allocator().start(
        curr_allocation, backend_commands, os.getcwd(), os.getcwd()
    )


@contextlib.contextmanager
def manage_workers(launch_workers):
    """Context manager for terminating workers on error."""
    workers = launch_workers()
    try:
        yield workers
    finally:
        workers.terminate(wait=0.3)


def backend_event_loop(run_feeder, server, ensemble_db, launch_workers, exit_condition):
    """The primary event loop driving top-level ensemble progress.

    :param launch_workers: a callable that submits workers that execute the runs
        of the ensemble
    :param exit_condition: a callable that returns True when it is safe to exit
    """
    while True:
        LOGGER.info("Entering new worker round")
        with manage_workers(launch_workers) as workers:
            inner_backend_event_loop(workers, run_feeder, server, ensemble_db)
        LOGGER.info("Worker round complete. Checking for overall completion...")
        if exit_condition() or not run_feeder.fill_cache():
            break
        LOGGER.info("Not yet complete.")
        run_feeder.fill_cache()


def inner_backend_event_loop(workers, run_feeder, server, ensemble_db):
    """An event loop contained within `backend_event_loop`"""
    curr_time = time.time()
    time_of_last_save = curr_time
    save_frequency = 1  # write the server's cache to the database every second
    while workers.poll() is None and server.is_alive():
        run_feeder.poll_cache()
        if curr_time - time_of_last_save > save_frequency:
            server.save_progress(ensemble_db)
        time.sleep(0.01)
        curr_time = time.time()
    server.save_progress(ensemble_db)
    if not server.is_alive():
        raise IOError("Server {!r} died".format(server))


def setup_ensemble(status_store, app_interface, root_dir):
    """Perform ensemble setup: call `prep_ensemble` if necessary."""
    prep_ensemble = getattr(app_interface, "prep_ensemble", None)
    if callable(prep_ensemble) and not status_store.fetch()["prepped"]:
        LOGGER.info("Calling user's prep_ensemble")
        backend.user_prep_ensemble(prep_ensemble, root_dir)
        status_store.set(prepped=True)
    else:
        LOGGER.info("No prep_ensemble to call. Proceeding...")


def can_exit(app_interface, status_store, max_failed_runs, root_dir, on_completion):
    """Return True if it is permissible to exit the submission loop."""
    status = status_store.fetch()
    if status["stop"]:
        LOGGER.info("'Stop ensemble' message received. Exiting...")
        return True
    if status["failed_runs"] >= max_failed_runs > -1:
        LOGGER.info("Max failed runs exceeded. Exiting...")
        return True
    post_ensemble = getattr(app_interface, "post_ensemble", None)
    exit_ok = True  # ok to exit unless post_ensemble or on_completion defined
    # post_ensemble and on_completion may add new runs
    if isinstance(on_completion, dict) and on_completion.get("args") is not None:
        backend.invoke_completion_script(on_completion)
        exit_ok = False
    if callable(post_ensemble):
        backend.user_post_ensemble(post_ensemble, root_dir)
        exit_ok = False
    if exit_ok:
        LOGGER.info("No post_ensemble/on_completion to call. Exiting...")
    return exit_ok


def use_ssl():
    """Create an SSL certificate for use by servers"""
    try:
        network.create_ssl_certificate()
    except OSError:
        LOGGER.exception("Couldn't generate SSL certificate:")


def early_stop_grace_period(runtime_end, grace_period_mins):
    """Go to sleep to keep an allocation alive without submitting any new runs."""
    grace_period_secs = grace_period_mins * 60
    grace_period_end = runtime_end + grace_period_secs
    remaining_grace_period_time = grace_period_end - time.time()
    if remaining_grace_period_time > 0:
        LOGGER.info("Entering grace period of %f seconds", remaining_grace_period_time)
        time.sleep(remaining_grace_period_time)


@contextlib.contextmanager
def allocation_manager(status_store, rmgr, runtime_end, args):
    """Upon catching a signal, enter the grace period and then create a new alloc."""
    try:
        yield
    except SignalCaught:
        early_stop_grace_period(runtime_end, args.early_stop)
        create_new_allocation(status_store, rmgr, args)


def setup_parser():
    """Setup the command-line parser for the script."""
    parser = argparse.ArgumentParser(
        description=(
            "Themis backend. Requires a database "
            "holding information about the ensemble."
        )
    )
    parser.add_argument(
        "-t",
        "--runtime",
        metavar="N",
        type=float,
        help="the max number of minutes to run for",
        required=True,
    )
    parser.add_argument(
        "-e",
        "--early-stop",
        metavar="N",
        type=float,
        help="the number of minutes early to stop executing runs",
        default=0,
    )
    parser.add_argument(
        "-a",
        "--alloc-id",
        metavar="N",
        type=int,
        help="integer ID identifying allocation in status store",
        required=True,
    )
    parser.add_argument(
        "-p",
        "--parallelism",
        metavar="N",
        type=int,
        help="the number of workers to launch",
        default=0,
        required=True,
    )
    parser.add_argument(
        "-c",
        "--max-concurrency",
        metavar="N",
        type=int,
        help="the maximum number of concurrent runs to allow",
        required=True,
    )
    parser.add_argument(
        "--setup-dir", help="Path to the Themis setup directory", default=os.curdir,
    )
    parser.add_argument(
        "--multiple",
        help=(
            "Allow multiple concurrent executions, disabling "
            "the automatic freeing of all queued runs"
        ),
        action="store_true",
    )
    return parser


def collect_args_set_timer():
    """Parse command-line arguments and set a runtime SIGALRM timer."""
    use_ssl()
    args = setup_parser().parse_args()
    args.setup_dir = os.path.abspath(args.setup_dir)
    LOGGER.info(
        "Running on %s for %f minutes, stopping run submission %i minutes early",
        socket.gethostname(),
        args.runtime,
        args.early_stop,
    )
    LOGGER.info("Running with parallelism %i", args.parallelism)
    if "FLUX_URI" in os.environ:
        LOGGER.info(
            "Flux remote URI is %s, local URI is %s",
            "ssh://"
            + socket.gethostname()
            + os.environ["FLUX_URI"].replace("local://", ""),
            os.environ["FLUX_URI"],
        )
    # catch signals to stop running, e.g. when the batch allocation is expiring
    proper_runtime_seconds = ((args.runtime - args.early_stop) * 60) - 5
    if proper_runtime_seconds >= 0:
        signal.alarm(int(proper_runtime_seconds))  # request an alarm signal
    backend.handle_signals()  # set signal handler
    runtime_end = time.time() + proper_runtime_seconds - 5
    return (args, runtime_end)


def main():
    """Set up and run the backend."""
    args, runtime_end = collect_args_set_timer()
    app_spec = database.get_app_spec(args.setup_dir)
    rmgr = resource.identify_resource_manager(
        app_spec["resource_mgr"], path=app_spec["flux_path"]
    )
    ensemble_db = database.default_database(app_spec)
    if not args.multiple:
        # since we are just starting up Themis, if we don't expect any other
        # concurrent backend processes, enable all runs
        ensemble_db.enable_all()
    app_interface = utils.import_app_interface(app_spec["app_interface"])
    status_store = database.get_status_store(args.setup_dir)
    # now export information and setup the objects used to run
    backend.export_to_user_utils(app_spec, status_store)
    setup_ensemble(status_store, app_interface, app_spec["root_dir"])
    with allocation_manager(status_store, rmgr, runtime_end, args), RunFeeder(
        ensemble_db, args.max_concurrency * 10, enable_runs=True
    ) as run_feeder, backend.managed_shutdown(
        network.BackendServer(network.CERT_NAME, run_feeder), ensemble_db
    ) as server:
        LOGGER.info("Set up server %s running on %s", server, server.server_address)
        backend_event_loop(
            run_feeder,
            server,
            ensemble_db,
            functools.partial(
                rmgr.launch_workers,
                args.parallelism,
                server.server_address,
                args.setup_dir,
                args.max_concurrency,
            ),
            functools.partial(
                can_exit,
                app_interface,
                status_store,
                app_spec["max_failed_runs"],
                app_spec["root_dir"],
                app_spec.get("on_completion"),
            ),
        )


def main_wrapper():
    """Setup the logging facility and wrap _main in a profiler."""
    with backend.managed_logging(sys.stderr):
        profilers = backend.profile_from_new(main)
        backend.dump_profile_data("backend_profile.pstat", *profilers)


if __name__ == "__main__":
    main_wrapper()
